"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module for wrapping style checks of the Magni source code.

"""

from __future__ import division
import contextlib
import os
import re
import unittest
try:
    from StringIO import StringIO  # Python 2 byte str (Python 2 only)
except ImportError:
    from io import StringIO  # Python 3 unicode str (both Py2 and Py3)

import pep8
import pyflakes.api
import radon.cli
import radon.complexity

import magni


class TestStyleConformance(unittest.TestCase):
    """
    Test of code conformance to style guides.

    """

    def setUp(self):
        """
        Identify magni source files.

        """

        path, name, ext = magni.utils.split_path(__file__)
        self.magni_base_path = path.rsplit(os.sep, 2)[0]
        self.exclude_patterns = ['__pycache__', 'tests']

        self.source_files = [os.path.join(dirpath, filename)
                             for dirpath, dirnames, filenames in
                             os.walk(self.magni_base_path)
                             for filename in filenames
                             if filename[-3:] == '.py' and
                             not any(
                                 [exclude_pattern in dirpath
                                  for exclude_pattern in
                                  self.exclude_patterns])]

    def test_pep8_conformance(self):
        """
        Tests for PEP8 conformance.

        Based on the example from the official documentation
        (http://pep8.readthedocs.org/en/latest/advanced.html)

        """

        style = pep8.StyleGuide()
        result = style.check_files(self.source_files)
        self.assertEqual(result.total_errors, 0)

    def test_pyflakes_conformance(self):
        """
        Tests for Python source code errors using Pyflakes

        """

        pyflakes_ignores = ['unable to detect undefined names',
                            'imported but unused',
                            'is assigned to but never used',
                            'redefinition of unused',
                            "undefined name 'xrange'"]
        ignore_re = re.compile('|'.join(pyflakes_ignores))

        with contextlib.closing(StringIO()) as out:
            with contextlib.closing(StringIO()) as err:

                reporter = pyflakes.reporter.Reporter(out, err)
                for file in self.source_files:
                    pyflakes.api.checkPath(file, reporter=reporter)

                out_value = out.getvalue()
                err_value = err.getvalue()

        out_filtered = [line for line in out_value.splitlines()
                        if not ignore_re.search(line)]

        self.assertEqual(len(err_value), 0,
                         msg='Pyflakes errors:\n {!s}'.format(err_value))

        self.assertEqual(len(out_filtered), 0,
                         msg='Pyflake warnings:\n {!s}'.format(out_filtered))

    def test_CC_conformance(self):
        """
        Tests for Cyclomatic Complexity levels below 10.

        *Corresponds to running from the magni folder*

        radon cc --total-average --no-assert --show-complexity --order SCORE
            --exclude "magni/utils/validation.py" --ignore "magni/tests*" magni

        """

        cc_threshold = 10
        exclude_modules = [os.path.join(
            'utils', 'validation', '_deprecated.py')]
        excludes = ','.join([os.path.join(self.magni_base_path, exclude)
                             for exclude in exclude_modules])

        config = radon.cli.Config(
            min='A',
            max='F',
            exclude=excludes,
            ignore='tests',
            show_complexity=True,
            average=False,
            total_average=True,
            order=radon.complexity.LINES,
            no_assert=True,
            show_closures=False)

        cc_harvester = radon.cli.CCHarvester([self.magni_base_path], config)

        cc_problems = dict()
        for key, val in cc_harvester._to_dicts().items():
            for k, element in enumerate(val):
                if element['complexity'] >= cc_threshold:
                    cc_problems['{} -- {}'.format(key, k)] = element

        cc_problems_print = [mod + ' :\n' + '\n'.join(
            [' : '.join([key, str(val)])
             for key, val in sorted(cc_problems[mod].items())])
            for mod in cc_problems]

        radon.cli.log_result(cc_harvester)
        self.assertFalse(
            cc_problems, msg=('Cyclomatic complexity exceeds {}:\n\n'
                              .format(cc_threshold) +
                              '\n\n'.join(cc_problems_print)))



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing configuration options for the phase_transition subpackage.

See also
--------
magni.utils.config.Configger : The Configger class used

Notes
-----
This module instantiates the `Configger` class provided by
`magni.utils.config`. The configuration options are the following:

algorithm_kwargs : dict
    The keyword arguments passed on to the reconstruction algorithm.
coefficients : {'rademacher', 'gaussian', 'laplace', 'bernoulli'}
    The distribution which the non-zero coefficients in the coefficient vector
    are drawn from.
custom_noise_factory : callable
    The callable that generates a custom m-by-1 vector of noise from the four
    arguments "m", "n", "k", "noise_power".
custom_system_matrix_factory : callable
    The callable that generates a custom m-by-n system matrix from the two
    arguments "m", "n".
delta : list or tuple
    The delta values of the delta-rho grid whose points are used for the monte
    carlo simulations (the default is [0., 1.]).
logit_solver : {'built-in', 'sklearn'}
    The solver to use in the logistic regression fit of the phase transition.
maxpoints : int
    The maximum number of phase space grid points to be handled by a process
    before it is replaced by a new process to free ressources (the default is
    None, which implies that processes are not replaced).
monte_carlo : int
    The number of monte carlo simulations to run in each point of the delta-rho
    grid (the default is 1).
noise : {'AWGN', 'AWLN', 'custom'}
    The type of noise to use (the default is None, which implies that
    noiseless measurements are used).
problem_size : int
    The length of the coefficient vector (the default is 800).
rho : list or tuple
    The rho values of the delta-rho grid whose points are used for the monte
    carlo simulations (the default is [0., 1.]).
seed : int
    The seed used when picking seeds for generating data for the monte carlo
    simulations or using the scikit-learn logistic regression solver (the
    default is None, which implies an arbitrary seed).
SNR : int or float
    The signal-to-noise ratio in decibel to use (the default is 40 dB).
support_distribution : ndarray
    The n x 1 support distribution array to use in the generation of test
    vectors (the default is None, which implies that the active entries in test
    vectors are the low index entries).
system_matrix : {'USE', 'RandomDCT2D', 'custom'}
    The system matrix to use in the simulation (the default is USE, which
    implies that the system matrix is drawn from the Uniform Spherical Ensemble
    ).

"""

from __future__ import division
from types import FunctionType as _FunctionType
try:
    from collections import Callable as _Callable  # Python 2
except ImportError:
    from collections.abc import Callable as _Callable  # Python 3

from magni.utils.config import Configger as _Configger
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric


configger = _Configger(
    {'algorithm_kwargs': {},
     'coefficients': 'rademacher',
     'custom_noise_factory': (
         lambda noise_power, m: 'Remember to configure custom noise factory'),
     'custom_system_matrix_factory': (
         lambda m, n: 'Remember to configure custom system matrix factory'),
     'delta': [0.0, 1.0],
     'logit_solver': 'built-in',
     'maxpoints': None,
     'monte_carlo': 1,
     'noise': None,
     'problem_size': 800,
     'rho': [0.0, 1.0],
     'seed': None,
     'SNR': 40,
     'support_distribution': None,
     'system_matrix': 'USE'},
    {'algorithm_kwargs': _generic(None, 'mapping'),
     'coefficients': _generic(
         None, 'string',
         value_in=('rademacher', 'gaussian', 'laplace', 'bernoulli')),
     'custom_noise_factory': _generic(
         None, (_FunctionType, _Callable)),
     'custom_system_matrix_factory': _generic(
         None, (_FunctionType, _Callable)),
     'delta': _levels(None, (
         _generic(None, 'explicit collection'),
         _numeric(None, 'floating', range_='[0;1]'))),
     'logit_solver': _generic(None, 'string',
                              value_in=('built-in', 'sklearn')),
     'maxpoints': _numeric(
         None, 'integer', range_='[1;inf)', ignore_none=True),
     'monte_carlo': _numeric(None, 'integer', range_='[1;inf)'),
     'noise': _generic(None, 'string',
                       value_in=('AWGN', 'AWLN', 'custom'), ignore_none=True),
     'problem_size': _numeric(None, 'integer', range_='[1;inf)'),
     'rho': _levels(None, (
         _generic(None, 'explicit collection'),
         _numeric(None, 'floating', range_='[0;1]'))),
     'seed': _numeric(None, 'integer', range_='[0;inf)', ignore_none=True),
     'SNR': _numeric(None, ('integer', 'floating')),
     'support_distribution': _numeric(None, 'floating', range_='[0;1]',
                                      shape=(-1, 1), ignore_none=True),
     'system_matrix': _generic(None, 'string',
                               value_in=('USE', 'RandomDCT2D', 'custom'))})



"""
..
    Copyright (c) 2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing unittests for `magni.utils.validation`.

"""

from __future__ import division
import unittest

import numpy as np

import magni
from magni.utils.validation import *


class AbstractTest(object):
    def _test_invalid_argument(self, args, kwargs, value):
        for kwarg in kwargs:
            def example(var):
                @decorate_validation
                def validate_input():
                    self._validate('var', *args, **kwarg)

                validate_input()

            behaves_correctly = False

            try:
                example(value)
            except Exception as e:
                if e.args[0][:24] == 'Invalid validation call:':
                    behaves_correctly = True
                else:
                    print('\nException caught:\n    {}'.format(e.args[0]))

            if not behaves_correctly:
                kwarg = list(kwarg.keys()) + list(kwarg.values())
                self.fail('{!r} should not accept kwarg, {}={!r}.'
                          .format(self._validate.__name__, *kwarg))

    def _test_invalid_value(self, args, kwarg, values, exc):
        def example(var):
            @decorate_validation
            def validate_input():
                self._validate('var', *args, **kwarg)

            validate_input()

        for value in values:
            behaves_correctly = False

            try:
                example(value)
            except exc as e:
                if e.args[0][:24] == 'Invalid validation call:':
                    raise
                else:
                    behaves_correctly = True

            if not behaves_correctly:
                self.fail('{!r} should fail here.'
                          .format(self._validate.__name__))

    def _test_valid_argument(self, args, kwargs, value):
        for kwarg in kwargs:
            def example(var):
                @decorate_validation
                def validate_input():
                    self._validate('var', *args, **kwarg)

                validate_input()

            behaves_correctly = True

            try:
                example(value)
            except Exception as e:
                if e.args[0][:24] == 'Invalid validation call:':
                    behaves_correctly = False
                else:
                    raise

            if not behaves_correctly:
                kwarg = list(kwarg.keys()) + list(kwarg.values())
                self.fail('{!r} should accept kwarg {}={!r}.'
                          .format(self._validate.__name__, *kwarg))

    def _test_valid_value(self, args, kwarg, values):
        def example(var):
            @decorate_validation
            def validate_input():
                self._validate('var', *args, **kwarg)

            validate_input()

        for value in values:
            try:
                example(value)
            except Exception as e:
                print('\nException caught:\n    {}'.format(e.args[0]))
                self.fail('{!r} should not fail here.'
                          .format(self._validate.__name__))


class TestCommon(unittest.TestCase):
    """
    Test of common validation function functionality.

    Implemented tests:

    * test_ignore_none : ensure that all validation functions can ignore none
      values if flagged to do so.
    * test_variable_passing : ensure that function arguments can be passed
      directly to validation functions to circumvent automatic function
      argument retrieval.
    * test_variable_retrieval : ensure that all validation functions can
      retrieve function arguments automatically.

    """

    _validators = (validate_generic, validate_numeric, validate_levels)

    def test_ignore_none(self):
        for validate in TestCommon._validators:
            if validate is validate_generic:
                arg = 'string'
            elif validate is validate_levels:
                continue
            elif validate is validate_numeric:
                arg = 'integer',
            else:
                raise ValueError('Unknown validator: {!r}'
                                 .format(validate.__name__))

            self._test_ignore_none(validate, arg)

    def test_variable_passing(self):
        for validate in TestCommon._validators:
            if validate is validate_generic:
                arg = 'string'
                var = ''
            elif validate is validate_levels:
                continue
            elif validate is validate_numeric:
                arg = 'integer',
                var = 0
            else:
                raise ValueError('Unknown validator: {!r}'
                                 .format(validate.__name__))

            self._test_variable_passing(validate, arg, var)

    def test_variable_retrieval(self):
        for validate in TestCommon._validators:
            if validate is validate_generic:
                arg = 'string'
                var = ''
            elif validate is validate_levels:
                arg = (validate_generic(None, 'string'),)
                var = ''
            elif validate is validate_numeric:
                arg = 'integer',
                var = 0
            else:
                raise ValueError('Unknown validator: {!r}'
                                 .format(validate.__name__))

            self._test_variable_retrieval(validate, arg, var)

    def _test_ignore_none(self, validate, arg):
        def example(var):
            @decorate_validation
            def validate_input():
                validate('var', arg)

            validate_input()

        def example_ignore(var):
            @decorate_validation
            def validate_input():
                validate('var', arg, ignore_none=True)

            validate_input()

        try:
            example(None)
        except ValueError as e:
            if e.args[0][:24] == 'Invalid validation call:':
                raise
        else:
            self.fail('{!r} should not blindly ignore {!r}.'
                      .format(validate.__name__, None))

        try:
            example_ignore(None)
        except ValueError as e:
            print('\nException caught:\n    {}'.format(e.args[0]))
            self.fail('{!r} should be able to ignore {!r}.'
                      .format(validate.__name__, None))

    def _test_variable_passing(self, validate, arg, var):
        def example(var):
            @decorate_validation
            def validate_input():
                validate('some_unlikely_name', arg, var=var)

            validate_input()

        try:
            example(var)
        except NameError as e:
            print('\nException caught:\n    {}'.format(e.args[0]))
            self.fail('{!r} should use passed function arguments.'
                      .format(decorate_validation.__name__))

    def _test_variable_retrieval(self, validate, arg, var):
        def example(var, var_in_sequence, var_in_mapping):
            @decorate_validation
            def validate_input():
                validate('var', arg)
                validate(('var_in_sequence', 0), arg)
                validate(('var_in_mapping', 'key'), arg)

            validate_input()

        def example_nonsense_name(var):
            @decorate_validation
            def validate_input():
                validate('no such var', arg)

            validate_input()

        def example_nonsense_index(var):
            @decorate_validation
            def validate_input():
                validate(('var', 'no such index'), arg)

            validate_input()

        try:
            example(var, [var], {'key': var})
        except (NameError, LookupError) as e:
            print('\nException caught:\n    {}'.format(e.args[0]))
            self.fail('{!r} should be able to locate the argument.'
                      .format(validate.__name__))

        try:
            example_nonsense_name(var)
        except NameError:
            pass
        else:
            self.fail('{!r} should not be able to locate this argument.'
                      .format(validate.__name__))

        try:
            example_nonsense_index(var)
        except LookupError:
            pass
        else:
            self.fail('{!r} should not be able to locate this argument.'
                      .format(validate.__name__))


class TestGeneric(unittest.TestCase, AbstractTest):
    """
    Test of validate_generic.

    Implemented tests:

    * test_has_keys : ensure that the function only accepts certain types for
      the has_keys argument, and that key checking works as intended.
    * test_keys_in : ensure that the function only accepts certain types for
      the keys_in argument, and that key checking works as intended.
    * test_len : ensure that the function only accepts certain types for the
      len argument, and that length checking works as intended.
    * test_superclass : ensure that the function only accepts certain types for
      the superclass argument, and that superclass checking works as intended.
    * test_type : ensure that the function only accepts certain types for the
      type_ argument, and that type checking works as intended.
    * test_value_in : ensure that the function only accepts certain types for
      the value_in argument, and that value checking works as intended.

    """

    def setUp(self):
        self._validate = validate_generic

    def test_has_keys(self):
        args = ('mapping',)

        value = {'key': 0}
        kwargs = ({'has_keys': ('key',)},)
        self._test_valid_argument(args, kwargs, value)

        kwargs = ({'has_keys': ''}, {'has_keys': 0})
        self._test_invalid_argument(args, kwargs, value)

        kwarg = {'has_keys': ('key',)}
        values = ({'key': 0, 'another_key': 0},)
        self._test_valid_value(args, kwarg, values)

        values = ({'unknown_key': 0, 'another_key': 0},)
        self._test_invalid_value(args, kwarg, values, KeyError)

    def test_keys_in(self):
        args = ('mapping',)

        value = {'key': 0}
        kwargs = ({'keys_in': ('key',)},)
        self._test_valid_argument(args, kwargs, value)

        kwargs = ({'keys_in': ''}, {'keys_in': 0})
        self._test_invalid_argument(args, kwargs, value)

        kwarg = {'keys_in': ('key', 'another_key')}
        values = ({'key': 0}, {'another_key': 0},)
        self._test_valid_value(args, kwarg, values)

        values = ({'unknown_key': 0, 'another_key': 0},)
        self._test_invalid_value(args, kwarg, values, KeyError)

    def test_len(self):
        args = ('explicit collection',)

        value = [0]
        kwargs = ({'len_': 1},)
        self._test_valid_argument(args, kwargs, value)

        kwargs = ({'len_': ''}, {'len_': ()})
        self._test_invalid_argument(args, kwargs, value)

        kwarg = {'len_': 3}
        values = ([0, 1, 2], (0, 1, 2))
        self._test_valid_value(args, kwarg, values)

        values = ([], (0, 1, 2, 3, 4))
        self._test_invalid_value(args, kwarg, values, ValueError)

    def test_superclass(self):
        args = ('class',)

        value = str
        kwargs = ({'superclass': object},)
        self._test_valid_argument(args, kwargs, value)

        kwargs = ({'superclass': object()}, {'superclass': (str, int)})
        self._test_invalid_argument(args, kwargs, value)

        kwarg = {'superclass': magni.utils.validation.types.MatrixBase}
        values = (magni.utils.matrices.Matrix,)
        self._test_valid_value(args, kwarg, values)

        values = (object, str)
        self._test_invalid_value(args, kwarg, values, TypeError)

    def test_type(self):
        args = ()

        value = ''
        kwargs = ({'type_': 'string'}, {'type_': str}, {'type_': (str, int)})
        self._test_valid_argument(args, kwargs, value)

        kwargs = ({'type_': 'unknown type'}, {'type_': 0})
        self._test_invalid_argument(args, kwargs, value)

        kwarg = {'type_': 'explicit collection'}
        values = ([], ())
        self._test_valid_value(args, kwarg, values)

        values = ('', 0, {})
        self._test_invalid_value(args, kwarg, values, TypeError)

    def test_value_in(self):
        args = ('string',)

        value = 'yes'
        kwargs = ({'value_in': ('yes',)},)
        self._test_valid_argument(args, kwargs, value)

        kwargs = ({'value_in': 0}, {'value_in': {}})
        self._test_invalid_argument(args, kwargs, value)

        kwarg = {'value_in': ('yes', 'no')}
        values = ('yes', 'no')
        self._test_valid_value(args, kwarg, values)

        values = ('maybe',)
        self._test_invalid_value(args, kwarg, values, ValueError)


class TestLevels(unittest.TestCase, AbstractTest):
    """
    Test of validate_levels.

    Implemented tests:

    * test_levels : ensure that the function only accepts valid level
      specifications for the levels argument, and that level checking works as
      intended.

    """

    def setUp(self):
        self._validate = validate_levels

    def test_levels(self):
        args = ()

        value = ('some', 'values')
        kwargs = (
            {'levels': (validate_generic(None, tuple),)},
            {'levels': (
                validate_generic(None, 'explicit collection'),
                validate_generic(None, 'string'))})
        self._test_valid_argument(args, kwargs, value)

        kwargs = (
            {'levels': ''},
            {'levels': ()},
            {'levels': ('',)},
            {'levels': ((),)},
            {'levels': (('levels',),)})
        self._test_invalid_argument(args, kwargs, value)

        kwarg = {'levels': (
            validate_generic(None, 'explicit collection'),
            validate_generic(None, 'string'))}
        values = (('some', 'values'), [], ['', '', ''])
        self._test_valid_value(args, kwarg, values)

        values = ({'key': 'value'}, (0, 1, 2), [{}, (), []])
        self._test_invalid_value(args, kwarg, values, TypeError)


class TestNumeric(unittest.TestCase, AbstractTest):
    """
    Test of validate_numeric.

    Implemented tests:

    * test_precision : ensure that the function only accepts known bit-lengths
      for the precision argument, and that precision checking works as
      intended.
    * test_range : ensure that the function only accepts valid string
      specifications for the range_ argument, and that range checking works as
      intended.
    * test_shape : ensure that the function only accepts valid types for the
      shape argument, and that shape chekcing works as intended.
    * test_type : ensure that the function only accepts certain types for the
      type_ argument, and that type checking works as intended.

    """

    def setUp(self):
        self._validate = validate_numeric

    def test_precision(self):
        args = ('integer',)

        value = np.int64()
        kwargs = ({'precision': 64}, {'precision': (32, 64)})
        self._test_valid_argument(args, kwargs, value)

        kwargs = ({'precision': ''}, {'precision': 42})
        self._test_invalid_argument(args, kwargs, value)

        kwarg = {'precision': (32, 64)}
        values = (np.int32(), np.int64())
        self._test_valid_value(args, kwarg, values)

        values = (np.int8(), np.int16())
        self._test_invalid_value(args, kwarg, values, TypeError)

    def test_range(self):
        args = ('integer',)

        value = 1
        kwargs = ({'range_': '[0;inf]'}, {'range_': '(0;inf)'},
                  {'range_': ']0;inf['}, {'range_': ('[0;inf]', '[0;0]')})
        self._test_valid_argument(args, kwargs, value)

        kwargs = ({'range_': ')0;inf('}, {'range_': '[0,inf]'},
                  {'range_': (0, 10)}, {'range_': ('[0;1]', '[0;1]', ['0;1'])})
        self._test_invalid_argument(args, kwargs, value)

        kwarg = {'range_': ']0;inf]'}
        values = (1, np.iinfo(np.int64).max)
        self._test_valid_value(args, kwarg, values)

        values = (0, np.iinfo(np.int64).min)
        self._test_invalid_value(args, kwarg, values, ValueError)

    def test_shape(self):
        args = ('floating',)

        value = np.zeros((5, 5))
        kwargs = ({'shape': (5, 5)}, {'shape': (5, -1)})
        self._test_valid_argument(args, kwargs, value)

        kwargs = ({'shape': ''}, {'shape': 0}, {'shape': ('', '')})
        self._test_invalid_argument(args, kwargs, value)

        kwarg = {'shape': (5, -1)}
        values = (np.zeros((5, 5)), np.zeros((5, 10)))
        self._test_valid_value(args, kwarg, values)

        values = (0., np.zeros((5, 5, 5)), np.zeros((10, 5)))
        self._test_invalid_value(args, kwarg, values, ValueError)

    def test_type(self):
        args = ()

        value = 0
        kwargs = ({'type_': 'integer'}, {'type_': ('integer', 'floating')})
        self._test_valid_argument(args, kwargs, value)

        kwargs = ({'type_': 'unknown type'}, {'type_': 0}, {'type_': int})
        self._test_invalid_argument(args, kwargs, value)

        kwarg = {'type_': ('integer', 'floating')}
        values = (0, 1., np.int8())
        self._test_valid_value(args, kwarg, values)

        values = ('', 1j, np.bool8())
        self._test_invalid_value(args, kwarg, values, TypeError)


class TestTypes(unittest.TestCase):
    """
    Test of custom validation types.

    Implemented tests:

    * testMMSEInputChannel : ensure that methods raise NotImplementedError
    * testMMSEOutputChannel : ensure that methods raise NotImplementedError
    * testStopCriterion : ensure that methods raise NotImplementedError
    * testTresholdOperator : ensure that methods raise NotImplementedError

    """

    def testMMSEInputChannel(self):
        channel = magni.utils.validation.types.MMSEInputChannel(dict())

        with self.assertRaises(NotImplementedError):
            channel.compute(dict())

    def testMMSEOutputChannel(self):
        channel = magni.utils.validation.types.MMSEOutputChannel(dict())

        with self.assertRaises(NotImplementedError):
            channel.compute(dict())

    def testStopCriterion(self):
        stop_criterion = magni.utils.validation.types.StopCriterion(dict())

        with self.assertRaises(NotImplementedError):
            stop_criterion.compute(dict())

    def testThresholdOperator(self):
        operator = magni.utils.validation.types.ThresholdOperator(dict())

        with self.assertRaises(NotImplementedError):
            operator.compute_deriv_threshold(dict())

        with self.assertRaises(NotImplementedError):
            operator.compute_threshold(dict())

        with self.assertRaises(NotImplementedError):
            operator.update_threshold_level(dict())


class TestUtils(unittest.TestCase):
    """
    Test of validation utility functions.

    Implemented tests:

    * test_disable_validation : ensure that all validation functions can be
      disabled.
    * test_require_decoration : ensure that all validation functions do not run
      if they are not decorated by decorate_validation.
    * test_validate_once : ensure that all validation functions run only once
      when validate_once is enabled.

    """

    _validators = (validate_generic, validate_numeric, validate_levels)

    def test_disable_validation(self):
        for validate in TestUtils._validators:
            if validate is validate_generic:
                arg = 'string'
                var = 0
            elif validate is validate_levels:
                arg = (validate_generic(None, 'string'),)
                var = 0
            elif validate is validate_numeric:
                arg = 'integer',
                var = ''
            else:
                raise ValueError('Unknown validator: {!r}'
                                 .format(validate.__name__))

            self._test_disable_validation(validate, arg, var)

    def test_require_decoration(self):
        for validate in TestUtils._validators:
            if validate is validate_generic:
                arg = 'string'
                var = ''
            elif validate is validate_levels:
                arg = (validate_generic(None, 'string'),)
                var = ''
            elif validate is validate_numeric:
                arg = 'integer',
                var = 0
            else:
                raise ValueError('Unknown validator: {!r}'
                                 .format(validate.__name__))

            self._test_require_decoration(validate, arg, var)

    def test_validate_once(self):
        for validate in TestUtils._validators:
            if validate is validate_generic:
                arg = 'string'
                vars_ = ('', 0)
            elif validate is validate_levels:
                arg = (validate_generic(None, 'string'),)
                vars_ = ('', 0)
            elif validate is validate_numeric:
                arg = 'integer',
                vars_ = (0, '')
            else:
                raise ValueError('Unknown validator: {!r}'
                                 .format(validate.__name__))

            self._test_validate_once(validate, arg, vars_)

    def _test_disable_validation(self, validate, arg, var):
        def example(var):
            @decorate_validation
            def validate_input():
                validate('var', arg)

            validate_input()

        try:
            example(var)
        except TypeError:
            pass
        else:
            self.fail('{!r} should fail here.'.format(validate.__name__))

        try:
            disable_validation()
            example(var)
        except TypeError as e:
            print('\nException caught:\n    {}'.format(e.args[0]))
            self.fail('{!r} should disable validation.'
                      .format(disable_validation.__name__))
        finally:
            magni.utils.validation._util._disabled = False

    def _test_require_decoration(self, validate, arg, var):
        def example_with_decoration(var):
            @decorate_validation
            def validate_input():
                validate('var', arg)

            validate_input()

        def example_without_decoration(var):
            def validate_input():
                validate('var', arg)

            validate_input()

        try:
            example_with_decoration(var)
        except RuntimeError as e:
            print('\nException caught:\n    {}'.format(e.args[0]))
            self.fail('{!r} should enable validation.'
                      .format(decorate_validation.__name__))

        try:
            example_without_decoration(var)
        except RuntimeError:
            pass
        else:
            self.fail('{!r} should require decoration.'
                      .format(validate.__name__))

    def _test_validate_once(self, validate, arg, vars_):
        @validate_once
        def example(var):
            @decorate_validation
            def validate_input():
                validate('var', arg)

            validate_input()

        try:
            example(vars_[0])
            example(vars_[1])
        except TypeError:
            pass
        else:
            self.fail('{!r} should fail here with {!r} disabled.'
                      .format(validate.__name__, _validate_once.__name__))

        try:
            enable_validate_once()
            example(vars_[0])
            example(vars_[1])
        except TypeError as e:
            print('\nException caught:\n    {}'.format(e.args[0]))
            self.fail('{!r} should enable validation once.'
                      .format(enable_validate_once.__name__))
        finally:
            magni.utils.validation._util._validate_once_enabled = False



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing implementations of Iterative Thresholding (IT).

Routine listings
----------------
config
    Configger providing configuration options for this subpackage.
run(y, A)
    Run the IT reconstruction algorithm.

Notes
-----
Implementations of Iterative Hard Thresholding (IHT) [1]_, [2]_ as well as
implementations of Iterative Soft Thresholding (IST) [3]_, [4]_ are available.
It is also possible to configure the subpackage to use a model based approach
as described in [5]_.

References
----------
.. [1] T. Blumensath and M.E. Davies, "Iterative Thresholding for Sparse
   Approximations", *Journal of Fourier Analysis and Applications*, vol. 14,
   pp. 629-654, Sep. 2008.
.. [2] T. Blumensath and M.E. Davies, "Normalized Iterative Hard Thresholding:
   Guaranteed Stability and Performance", *IEEE Journal Selected Topics in
   Signal Processing*, vol. 4, no. 2, pp. 298-309, Apr. 2010.
.. [3] I. Daubechies, M. Defrise, and C. D. Mol, "An Iterative Thresholding
   Algorithm for Linear Inverse Problems with a Sparsity Constraint",
   *Communications on Pure and Applied Mathematics*, vol. 57, no. 11,
   pp. 1413-1457, Nov. 2004.
.. [4] A. Maleki and D.L. Donoho, "Optimally Tuned Iterative Reconstruction
   Algorithms for Compressed Sensing", *IEEE Journal Selected Topics in Signal
   Processing*, vol. 3, no. 2, pp. 330-341, Apr. 2010.
.. [5] R.G. Baraniuk, V. Cevher, M.F. Duarte, and C. Hedge, "Model-Based
   Compressive Sensing", *IEEE Transactions on Information Theory*, vol. 56,
   no. 4, pp. 1982-2001, Apr. 2010.

"""

from magni.cs.reconstruction.it._config import configger as config
from magni.cs.reconstruction.it._algorithm import run



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality for converting a file stream.

The reading of an .mi file is logically separated into four steps of which the
functionality provided by this module performs the first step.

Routine listings
----------------
convert_stream(stream)
    Convert a file stream to flat, basic variables.

"""

from __future__ import division

import numpy as np


def convert_stream(stream):
    """
    Convert a file stream to flat, basic variables.

    The flat, basic variables are separated into header parameters and the
    data. The header parameters are name, value pairs whereas the data is a
    1-dimensional numeric array.

    Parameters
    ----------
    stream : str
        The file stream.

    Returns
    -------
    params : tuple
        The header parameters.
    data : numpy.ndarray
        The data.

    See Also
    --------
    magni.afm.io.read_mi_file : Function using the present function.

    Notes
    -----
    This function splits the file stream into header and data, and splits the
    header into parameter name, value pairs.

    """

    index = stream.find(b'data          ')
    index = stream.find(b'\n', index)
    stream_params = stream[:index]
    stream_data = stream[index + 1:]

    params = stream_params.split(b'\n')
    params = [(str(param[:14].decode()), str(param[14:].decode()))
              for param in params]
    params = [(name.strip(), _convert_parameter_value(value))
              for name, value in params]
    params = tuple(params)

    if params[0][0] != 'fileType':
        raise IOError("The file must have a 'fileType' header parameter.")

    if params[-1][0] != 'data':
        raise IOError("The file must have a 'data' header parameter.")

    if params[0][1].lower() == 'image':
        data = _convert_image_data(stream_data, params[-1][1])
    elif params[0][1].lower() == 'spectroscopy':
        data = _convert_spectroscopy_data(stream_data, params[-1][1])
    else:
        msg = 'The value of  >>fileType<<, {!r}, must be in {!r}.'
        raise IOError(msg.format(params[0][1], ('Image', 'Spectroscopy')))

    return params, data


def _convert_image_data(buffer_, data_type):
    """
    Convert the file stream image data to a 1-dimensional numeric array.

    Parameters
    ----------
    buffer_ : str
        The file stream image data.
    data_type : str
        The data type of the file stream image data.

    Returns
    -------
    data : numpy.ndarray
        The data as a 1-dimensional numeric array.

    See Also
    --------
    convert_stream : Function using the present function.

    """

    if data_type.lower() == 'binary':
        data = np.frombuffer(buffer_, np.int16)
        data = np.float64(data) / 2**15
    elif data_type.lower() == 'binary_32':
        data = np.frombuffer(buffer_, np.int32)
        data = np.float64(data) / 2**31
    elif data_type.lower() == 'ascii':
        data = map(int, buffer_.strip().split())
        data = np.float64(data) / 2**15
    elif data_type.lower() == 'ascii_absolute':
        data = map(float, buffer_.strip().split())
        data = np.float64(data)
    elif data_type.lower() == 'ascii_multicolumn':
        data = buffer_.strip().split()
        data = [map(float, row.split()) for row in data]
        data = np.float64(data).flatten('F')
    else:
        raise IOError('Unknown data type {!r}.'.format(data_type))

    return data


def _convert_parameter_value(string):
    """
    Convert a file stream parameter value to a basic python value.

    The converted value is either a boolean, string, floating point, or integer
    value or a list containing a mix of these.

    Parameters
    ----------
    string : str
        The file stream parameter value.

    Returns
    -------
    value : None
        The converted, basic python value.

    See Also
    --------
    convert_stream : Function using the present function.

    """

    parts = string.split()

    if len(parts) == 0:
        value = None
    elif len(parts) == 1:
        if string.lower() == 'true':
            value = True
        elif string.lower() == 'false':
            value = False
        else:
            try:
                value = string
                value = float(string)
                value = int(string)
            except ValueError:
                pass
    else:
        value = [_convert_parameter_value(part) for part in parts]
        value = tuple(value) if not isinstance(value[0], str) else string

    return value


def _convert_spectroscopy_data(buffer_, data_type):
    """
    Convert the file stream spectroscopy data to a 1-dimensional numeric array.

    Parameters
    ----------
    buffer_ : str
        The file stream spectroscopy data.
    data_type : str
        The data type of the file stream spectroscopy data.

    Returns
    -------
    data : numpy.ndarray
        The data as a 1-dimensional numeric array.

    See Also
    --------
    convert_stream : Function using the present function.

    """

    if data_type.lower() == 'binary':
        data = np.frombuffer(buffer_, np.float32)
        data = np.float64(data)
    elif data_type.lower() == 'ascii_multicolumn':
        data = buffer_.strip().split()[1:]
        data = [map(float, row.split()[2:]) for row in data]
        data = np.float64(data).flatten('F')
    else:
        raise IOError('Unknown data type {!r}.'.format(data_type))

    return data



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing validation capability.

The intention is to validate all public functions of the package such that
erroneous arguments in calls are reported in an informative fashion rather than
causing arbitrary exceptions or unexpected results. To avoid performance
impairments, the validation can be disabled globally.

Routine listings
----------------
types
    Module providing abstract superclasses for validation.
decorate_validation(func)
    Decorate a validation function (see Notes).
disable_validation()
    Disable validation globally (see Notes).
enable_validate_once()
    Enable validating inputs only once (see Notes).
validate_generic(name, type, value_in=None, len_=None, keys_in=None,
    has_keys=None, superclass=None, ignore_none=False, var=None)
    Validate non-numeric objects.
validate_levels(name, levels)
    Validate containers and mappings as well as contained objects.
validate_numeric(name, type, range_='[-inf;inf]', shape=(), precision=None,
    ignore_none=False, var=None)
    Validate numeric objects.
validate_once(func)
    Decorate a function to allow for a one-time input validation (see Notes).

Notes
-----
To be able to disable validation (and to ensure consistency), every public
function or method should define a nested validation function with the name
'validate_input' which takes no arguments. This function should be decorated by
`decorate_validation`, be placed in the beginning of the parent function or
method, and be called as the first thing after its definition.

Functions in magni may be decorated by `validate_once`. If the validate once
functionality is enabled, these functions only validate their input arguments
on the first call to the function.

Examples
--------
If, for example, the following function is defined:

>>> def greet(person, greeting):
...     print('{}, {} {}.'.format(greeting, person['title'], person['name']))

This function expects its argument, 'person' to be a dictionary with keys
'title' and 'name' and its argument, 'greeting' to be a string. If, for
example, a list is passed as the first argument, a TypeError is raised with the
description 'list indices must be integers, not str'. While obviously correct,
this message is not excessively informative to the user of the function.
Instead, this module can be used to redefine the function as follows:

>>> from magni.utils.validation import decorate_validation, validate_generic
>>> def greet(person, greeting):
...     @decorate_validation
...     def validate_input():
...         validate_generic('person', 'mapping', has_keys=('title', 'name'))
...         validate_generic('greeting', 'string')
...     validate_input()
...     print('{}, {} {}.'.format(greeting, person['title'], person['name']))

If, again, a list is passed as the first argument, a TypeError with the
description "The value(s) of >>type(person)<<, <type 'list'>, must be in
('mapping',)." is raised. Now, the user of the function can easily identify the
mistake and correct the call to read:

>>> greet({'title': 'Mr.', 'name': 'Anderson'}, 'You look surprised to see me')
You look surprised to see me, Mr. Anderson.

"""

from magni.utils.validation import types
from magni.utils.validation._generic import validate_generic
from magni.utils.validation._numeric import validate_numeric
from magni.utils.validation._levels import validate_levels
from magni.utils.validation._util import decorate_validation
from magni.utils.validation._util import enable_validate_once
from magni.utils.validation._util import disable_validation
from magni.utils.validation._util import validate_once



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing backup capabilities for the monte carlo simulations.

The backup stores the simulation results and the simulation timings pointwise
for the points in the delta-rho simulation grid. The set function targets a
specific point while the get function targets the entire grid in order to keep
the overhead low.

Routine listings
----------------
create(path)
    Create the HDF5 backup database with the required arrays.
get(path)
    Return which of the results have been stored.
set(path, ij_tuple, stat_time, stat_dist)
    Store the simulation data of a specified point.

See Also
--------
magni.cs.phase_transition.config : Configuration options.

Notes
-----
In practice, the backup database includes an additional array for tracking for
which points data has been stored. By first storing the data and then modifying
this array, the data is guaranteed to have been stored, when the array is
modified.

"""

from __future__ import division
import os

import numpy as np

from magni.cs.phase_transition import config as _conf
from magni.utils.multiprocessing import File as _File


def create(path):
    """
    Create the HDF5 backup database with the required arrays.

    The required arrays are an array for the simulation results, an array for
    the simulation timings, and an array for tracking the status.

    Parameters
    ----------
    path : str
        The path of the HDF5 backup database.

    See Also
    --------
    magni.cs.phase_transition.config : Configuration options.

    """

    shape = [_conf[key] for key in ['delta', 'rho', 'monte_carlo']]
    shape = [len(shape[0]), len(shape[1]), shape[2]]
    time = dist = mse = norm = np.zeros(shape)
    status = np.zeros(shape[:2], np.bool8)

    try:
        with _File(path, 'w') as f:
            f.create_array('/', 'time', time)
            f.create_array('/', 'dist', dist)
            f.create_array('/', 'mse', mse)
            f.create_array('/', 'norm', norm)
            f.create_array('/', 'status', status)
    except BaseException as e:
        if os.path.exists(path):
            os.remove(path)

        raise e


def get(path):
    """
    Return which of the results have been stored.

    The returned value is a copy of the boolean status array indicating which
    of the results have been stored.

    Parameters
    ----------
    path : str
        The path of the HDF5 backup database.

    Returns
    -------
    status : ndarray
        The boolean status array.

    """

    with _File(path, 'r') as f:
        done = f.root.status[:]

    return done


def set(path, ij_tuple, stat_time, stat_dist, stat_mse, stat_norm):
    """
    Store the simulation data of a specified point.

    Parameters
    ----------
    path : str
        The path of the HDF5 backup database.
    ij_tuple : tuple
        A tuple (i, j) containing the parameters i, j as listed below.
    i : int
        The delta-index of the point in the delta-rho grid.
    j : int
        The rho-index of the point in the delta-rho grid.
    stat_dist : ndarray
        The simulation results of the specified point.
    stat_time : ndarray
        The simulation timings of the specified point.
    stat_mse : ndarray
        The simulation mean squared error of the specified point.
    stat_norm : ndarray
        The simulation true vector 2-norm.

    """

    i, j = ij_tuple

    with _File(path, 'a') as f:
        f.root.time[i, j] = stat_time
        f.root.dist[i, j] = stat_dist
        f.root.mse[i, j] = stat_mse
        f.root.norm[i, j] = stat_norm
        f.root.status[i, j] = True



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the public class of the magni.utils.multiprocessing
subpackage.

"""

from __future__ import division
import multiprocessing as mp

import tables

from magni.utils.validation import decorate_validation as _decorate_validation


_lock = mp.Lock()


class File():
    """
    Control pytables access to hdf5 files when using multiprocessing.

    `File` retains the interface of `tables.open_file` and should only be used
    in 'with' statements (see Examples).

    Parameters
    ----------
    args : tuple
        The arguments that are passed to 'tables.open_file'.
    kwargs : dict
        The keyword arguments that are passed to 'tables.open_file'.

    See Also
    --------
    tables.open_file : The wrapped function.

    Notes
    -----
    Internally the module uses a global lock which is shared amongst all files.
    This solution is simple and does not entail significant overhead. However,
    the wait time introduced when using multiple files at the same time can be
    significant.

    Examples
    --------
    The class is used in the following way:

    >>> from magni.utils.multiprocessing._util import File
    >>> with File('database.hdf5', 'a') as f:
    ...     pass # execute something involving the opened file

    """

    def __init__(self, *args, **kwargs):
        @_decorate_validation
        def validate_input():
            if len(args) == 0 and 'filename' not in kwargs:
                raise ValueError('File must be called with a filename '
                                 'argument.')

        validate_input()

        self._args = args
        self._kwargs = kwargs

        if len(args) > 0:
            self._filename = args[0]
        else:
            self._filename = kwargs['filename']

    def __enter__(self):
        """
        Acquire the global lock before opening and returning the file.

        Returns
        -------
        file : tables.File
            The file specified in the call to `__init__`.

        """

        _lock.acquire()
        self._file = tables.open_file(*self._args, **self._kwargs)
        return self._file

    def __exit__(self, type, value, traceback):
        """
        Release the global lock after closing the file.

        Parameters
        ----------
        type : type
            The type of the exception raised, if any.
        value : Exception
            The exception rasied, if any.
        traceback : traceback
            The traceback of the exception raised, if any.

        """

        self._file.close()
        _lock.release()



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing generic compressed sensing functionality.

Routine listings
----------------
indicators
    Module prodiving performance indicator determination functionality.
phase_transition
    Subpackage providing phase transition determination functionality.
reconstruction
    Subpackage providing implementations of generic reconstruction algorithms.

"""

from magni.cs import indicators
from magni.cs import phase_transition
from magni.cs import reconstruction



"""
..
    Copyright (c) 2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module prodiving performance indicator determination functionality.

Routine listings
----------------
calculate_coherence(Phi, Psi, norm=None)
    Calculate the coherence of the Phi Psi matrix product.
calculate_mutual_coherence(Phi, Psi, norm=None)
    Calculate the mutual coherence of Phi and Psi.
calculate_relative_energy(Phi, Psi, method=None)
    Calculate the relative energy of Phi Psi matrix product atoms.

Notes
-----
For examples of uses of the performance indicators, see the related papers on
predicting/modelling reconstruction quality [1]_, [2]_.

References
----------
.. [1] P. S. Pedersen, J. Ostergaard, and T. Larsen, "Predicting Reconstruction
   Quality within Compressive Sensing for Atomic Force Microscopy," *2015 IEEE
   Global Conference on Signal and Information Processing (GlobalSIP)*,
   Orlando, FL, 2015, pp. 418-422. doi: 10.1109/GlobalSIP.2015.7418229
.. [2] P. S. Pedersen, J. Ostergaard, and T. Larsen, "Modelling Reconstruction
   Quality of Lissajous Undersampled Atomic Force Microscopy Images," *2016
   IEEE 13th International Symposium on Biomedical Imaging (ISBI)*, Prague,
   Czech Republic, 2016, pp. 245-248. doi: 10.1109/ISBI.2016.7493255

"""

from __future__ import division

import numpy as np

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric
from magni.utils.validation import validate_generic as _generic


def calculate_coherence(Phi, Psi, norm=None):
    r"""
    Calculate the coherence of the Phi Psi matrix product.

    In the context of Compressive Sensing, coherence usually refers to the
    maximum absolute correlation between two columns of the Phi Psi matrix
    product. This function allows the usage of a different normalised norm
    where the infinity-norm yields the usual case.

    Parameters
    ----------
    Phi : magni.utils.matrices.Matrix or numpy.ndarray
        The measurement matrix.
    Psi : magni.utils.matrices.Matrix or numpy.ndarray
        The dictionary matrix.
    norm : int or float
        The normalised norm used for the calculation (the default value is None
        which implies that the 0-, 1-, 2-, and infinity-norms are returned).

    Returns
    -------
    coherence : float or dict
        The coherence value(s).

    Notes
    -----
    If `norm` is None, the function returns a dict containing the coherence
    using the 0-, 1-, 2-, and infinity-norms. Otherwise, the function returns
    the coherence using the specified norm.

    The coherence is calculated as:

    .. math::

        \left(\frac{1}{n^2 - n}
        \sum_{i = 1}^n \sum_{\substack{j = 1 \\ j \neq i}}^n \left(
        \frac{|\Psi_{:, i}^T \Phi^T \Phi \Psi_{:, j}|}
        {||\Phi \Psi_{:, i}||_2 ||\Phi \Psi_{:, j}||_2}
        \right)^{\text{norm}}\right)^{\frac{1}{\text{norm}}}

    where `n` is the number of columns in `Psi`. In the case of the 0-norm,
    the coherence is calculated as:

    .. math::

        \frac{1}{n^2 - n}
        \sum_{i = 1}^n \sum_{\substack{j = 1 \\ j \neq i}}^n \mathbf{1}
        \left(\frac{|\Psi_{:, i}^T \Phi^T \Phi \Psi_{:, j}|}
        {||\Phi \Psi_{:, i}||_2 ||\Phi \Psi_{:, j}||_2}\right)

    where :math:`\mathbf{1}(a)` is 1 if `a` is non-zero and 0 otherwise. In the
    case of the infinity-norm, the coherence is calculated as:

    .. math::

        \max_{\substack{i, j \in \{1, \dots, n\} \\ i \neq j}}
        \left(\frac{|\Psi_{:, i}^T \Phi^T \Phi \Psi_{:, j}|}
        {||\Phi \Psi_{:, i}||_2 ||\Phi \Psi_{:, j}||_2}\right)

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> import magni
    >>> from magni.cs.indicators import calculate_coherence
    >>> Phi = np.zeros((5, 9))
    >>> Phi[0, 0] = Phi[1, 2] = Phi[2, 4] = Phi[3, 6] = Phi[4, 8] = 1
    >>> Psi = magni.imaging.dictionaries.get_DCT((3, 3))
    >>> for item in sorted(calculate_coherence(Phi, Psi).items()):
    ...     print('{}-norm: {:.3f}'.format(*item))
    0-norm: 0.222
    1-norm: 0.141
    2-norm: 0.335
    inf-norm: 1.000

    The above values can be calculated individually by specifying a norm:

    >>> for norm in (0, 1, 2, np.inf):
    ...     value = calculate_coherence(Phi, Psi, norm=norm)
    ...     print('{}-norm: {:.3f}'.format(norm, value))
    0-norm: 0.222
    1-norm: 0.141
    2-norm: 0.335
    inf-norm: 1.000

    """

    @_decorate_validation
    def validate_input():
        _numeric('Phi', ('integer', 'floating', 'complex'), shape=(-1, -1))
        _numeric('Psi', ('integer', 'floating', 'complex'),
                 shape=(Phi.shape[1], -1))
        _numeric('norm', ('integer', 'floating'), range_='[0;inf]',
                 ignore_none=True)

    validate_input()

    A = np.zeros((Phi.shape[0], Psi.shape[1]))
    e = np.zeros((A.shape[1], 1))

    for i in range(A.shape[1]):
        e[i] = 1
        A[:, i] = Phi.dot(Psi.dot(e)).reshape(-1)
        e[i] = 0

    M = np.zeros((A.shape[1], A.shape[1]))
    PhiT = Phi.T
    PsiT = Psi.T

    for i in range(A.shape[1]):
        M[:, i] = np.abs(PsiT.dot(PhiT.dot(
            A[:, i].reshape(-1, 1)))).reshape(-1)
        M[i, i] = 0

    w = 1 / np.linalg.norm(A, axis=0).reshape(-1, 1)
    M = M * w * w.T

    if norm is None:
        entries = (M.size - M.shape[0])
        value = {0: np.sum(M > 1e-9) / entries,
                 1: np.sum(M) / entries,
                 2: (np.sum(M**2) / entries)**(1 / 2),
                 np.inf: np.max(M)}
    elif norm == 0:
        value = np.sum(M > 1e-9) / (M.size - M.shape[0])
    elif norm == np.inf:
        value = np.max(M)
    else:
        value = (np.sum(M**norm) / (M.size - M.shape[0]))**(1 / norm)

    return value


def calculate_mutual_coherence(Phi, Psi, norm=None):
    r"""
    Calculate the mutual coherence of Phi and Psi.

    In the context of Compressive Sensing, mutual coherence usually refers to
    the maximum absolute correlation between two columns of Phi and Psi. This
    function allows the usage of a different normalised norm where the
    infinity-norm yields the usual case.

    Parameters
    ----------
    Phi : magni.utils.matrices.Matrix or numpy.ndarray
        The measurement matrix.
    Psi : magni.utils.matrices.Matrix or numpy.ndarray
        The dictionary matrix.
    norm : int or float
        The normalised norm used for the calculation (the default value is None
        which implies that the 0-, 1-, 2-, and infinity-norms are returned).

    Returns
    -------
    mutual_coherence : float or dict
        The mutual_coherence value(s).

    Notes
    -----
    If `norm` is None, the function returns a dict containing the mutual
    coherence using the 0-, 1-, 2-, and infinity-norms. Otherwise, the function
    returns the mutual coherence using the specified norm.

    The mutual coherence is calculated as:

    .. math::

        \left(\frac{1}{m n} \sum_{i = 1}^m \sum_{j = 1}^n
        |\Phi_{i, :} \Psi_{:, j}|^{\text{norm}}\right)^{\frac{1}{\text{norm}}}

    where `m` is the number of rows in `Phi` and `n` is the number of columns
    in `Psi`. In the case of the 0-norm, the mutual coherence is calculated as:

    .. math::

        \frac{1}{m n} \sum_{i = 1}^m \sum_{j = 1}^n \mathbf{1}
        (|\Phi_{i, :} \Psi_{:, j}|)

    where :math:`\mathbf{1}(a)` is 1 if `a` is non-zero and 0 otherwise. In the
    case of the infinity-norm, the mutual coherence is calculated as:

    .. math::

        \max_{i \in \{1, \dots, m\}, j \in \{1, \dots, n\}}
        |\Phi_{i, :} \Psi_{:, j}|

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> import magni
    >>> from magni.cs.indicators import calculate_mutual_coherence
    >>> Phi = np.zeros((5, 9))
    >>> Phi[0, 0] = Phi[1, 2] = Phi[2, 4] = Phi[3, 6] = Phi[4, 8] = 1
    >>> Psi = magni.imaging.dictionaries.get_DCT((3, 3))
    >>> for item in sorted(calculate_mutual_coherence(Phi, Psi).items()):
    ...     print('{}-norm: {:.3f}'.format(*item))
    0-norm: 0.889
    1-norm: 0.298
    2-norm: 0.333
    inf-norm: 0.667

    The above values can be calculated individually by specifying a norm:

    >>> for norm in (0, 1, 2, np.inf):
    ...     value = calculate_mutual_coherence(Phi, Psi, norm=norm)
    ...     print('{}-norm: {:.3f}'.format(norm, value))
    0-norm: 0.889
    1-norm: 0.298
    2-norm: 0.333
    inf-norm: 0.667

    """

    @_decorate_validation
    def validate_input():
        _numeric('Phi', ('integer', 'floating', 'complex'), shape=(-1, -1))
        _numeric('Psi', ('integer', 'floating', 'complex'),
                 shape=(Phi.shape[1], -1))
        _numeric('norm', ('integer', 'floating'), range_='[0;inf]',
                 ignore_none=True)

    validate_input()

    M = np.zeros((Phi.shape[0], Psi.shape[1]))
    e = np.zeros((Psi.shape[1], 1))

    for i in range(M.shape[1]):
        e[i] = 1
        M[:, i] = np.abs(Phi.dot(Psi.dot(e))).reshape(-1)
        e[i] = 0

    if norm is None:
        value = {0: np.sum(M > 1e-9) / M.size,
                 1: np.sum(M) / M.size,
                 2: (np.sum(M**2) / M.size)**(1 / 2),
                 np.inf: np.max(M)}
    elif norm == 0:
        value = np.sum(M > 1e-9) / M.size
    elif norm == np.inf:
        value = np.max(M)
    else:
        value = (np.sum(M**norm) / M.size)**(1 / norm)

    return value


def calculate_relative_energy(Phi, Psi, method=None):
    r"""
    Calculate the relative energy of Phi Psi matrix product atoms.

    Parameters
    ----------
    Phi : magni.utils.matrices.Matrix or numpy.ndarray
        The measurement matrix.
    Psi : magni.utils.matrices.Matrix or numpy.ndarray
        The dictionary matrix.
    method : str
        The method used for summarising the relative energies of the Phi Psi
        matrix product atoms.

    Returns
    -------
    relative_energy : float or dict
        The relative_energy summary value(s).

    Notes
    -----
    The summary `method` used is either 'mean' for mean value, 'std' for
    standard deviation, 'min' for minimum value, or 'diff' for difference
    between the maximum and minimum values.

    If `method` is None, the function returns a dict containing all of the
    above summaries. Otherwise, the function returns the specified summary.

    The relative energies, which are summarised by the given `method`, are
    calculated as:

    .. math::

        \left[
        \frac{||\Phi \Psi_{:, 1}||_2}{||\Psi_{:, 1}||_2},
        \dots,
        \frac{||\Phi \Psi_{:, n}||_2}{||\Psi_{:, n}||_2}
        \right]^T

    where `n` is the number of columns in `Psi`.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> import magni
    >>> from magni.cs.indicators import calculate_relative_energy
    >>> Phi = np.zeros((5, 9))
    >>> Phi[0, 0] = Phi[1, 2] = Phi[2, 4] = Phi[3, 6] = Phi[4, 8] = 1
    >>> Psi = magni.imaging.dictionaries.get_DCT((3, 3))
    >>> for item in sorted(calculate_relative_energy(Phi, Psi).items()):
    ...     print('{}: {:.3f}'.format(*item))
    diff: 0.423
    mean: 0.735
    min: 0.577
    std: 0.126

    The above values can be calculated individually by specifying a norm:

    >>> for method in ('mean', 'std', 'min', 'diff'):
    ...     value = calculate_relative_energy(Phi, Psi, method=method)
    ...     print('{}: {:.3f}'.format(method, value))
    mean: 0.735
    std: 0.126
    min: 0.577
    diff: 0.423

    """

    @_decorate_validation
    def validate_input():
        _numeric('Phi', ('integer', 'floating', 'complex'), shape=(-1, -1))
        _numeric('Psi', ('integer', 'floating', 'complex'),
                 shape=(Phi.shape[1], -1))
        _generic('method', 'string', value_in=('min', 'diff', 'mean', 'std'),
                 ignore_none=True)

    validate_input()

    e = np.zeros((Psi.shape[1], 1))
    x = np.zeros((Psi.shape[0], 1), dtype=Psi.dtype)
    w = np.zeros(Psi.shape[1])

    for i in range(Psi.shape[1]):
        e[i] = 1
        x[:] = Psi.dot(e)
        w[i] = np.linalg.norm(Phi.dot(x)) / np.linalg.norm(x)
        e[i] = 0

    if method is None:
        value = {'min': np.min(w),
                 'diff': np.max(w) - np.min(w),
                 'mean': np.mean(w),
                 'std': np.std(w)}
    elif method == 'min':
        value = np.min(w)
    elif method == 'diff':
        value = np.max(w) - np.min(w)
    elif method == 'mean':
        value = np.mean(w)
    elif method == 'std':
        value = np.std(w)

    return value



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing atomic force miscroscopy specific functionality.

The present subpackage includes functionality for handling AFM files and data
and functionality for utilizing the other subpackages for such AFM data.

Routine listings
----------------
config
    Configger providing configuration options for this subpackage.
io
    Subpackage providing input/output functionality for .mi files.
reconstruction
    Module providing reconstruction and analysis of reconstructed images.
types
    Subpackage providing data container classes for .mi files.

"""

from magni.afm._config import configger as config
from magni.afm import io
from magni.afm import reconstruction
from magni.afm import types



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing functionality for constructing and visualising scan
patterns for measurements.

This subpackage provides several pairs of scan pattern functions. The first
function, named \*_sample_surface, is used for sampling a given surface. The
second function, named \*_sample_image, is a wrapper that provides a
pixel-oriented interface to the first function. In addition to these pairs of
scan pattern functions, the module provides auxillary functions that may be
used to visualise the scan patterns.

Routine listings
----------------
construct_measurement_matrix(coords, h, w)
    Function for constructing a measurement matrix.
construct_pixel_mask(h, w, pixels)
    Construct a binary pixel mask.
lissajous_sample_image(h, w, scan_length, num_points, f_y=1., f_x=1.,
    theta_y=0., theta_x=np.pi / 2)
    Function for lissajous sampling an image.
lissajous_sample_surface(l, w, speed, sample_rate, time, f_y=1., f_x=1.,
    theta_y=0., theta_x=np.pi / 2, speed_mode=0)
    Function for lissajous sampling a surface.
plot_pattern(l, w, coords, mode, output_path=None)
    Function for visualising a scan pattern.
plot_pixel_mask(h, w, pixels, output_path=None)
    Function for visualising a pixel mask obtained from a scan pattern.
random_line_sample_image(h, w, scan_length, num_points, discrete=None,
    seed=None)
    Function for random line sampling an image.
random_line_sample_surface(l, w, speed, sample_rate, time, discrete=None,
    seed=None)
    Function for random line sampling a surface.
spiral_sample_image(h, w, scan_length, num_points, rect_area=False)
    Function for spiral sampling an image.
spiral_sample_surface(l, w, speed, sample_rate, time, rect_area=False)
    Function for spiral sampling a surface.
square_spiral_sample_image(h, w, scan_length, num_points)
    Function for square spiral sampling an image.
square_spiral_sample_surface(l, w, speed, sample_rate, time)
    Function for square spiral sampling a surface.
uniform_line_sample_image(h, w, scan_length, num_points)
    Function for uniform line sampling an image.
uniform_line_sample_surface(l, w, speed, sample_rate, time)
    Function for uniform line sampling a surface.
uniform_rotated_line_sample_image(h, w, scan_length, num_points, angle=0.,
    follow_edge=True)
    Function for uniform rotated line sampling an image.
uniform_rotated_line_sample_surface(l, w, speed, sample_rate, time, angle=0.,
    follow_edge=True)
    Function for uniform rotated line sampling a surface.
unique_pixels(coords)
    Function for determining unique pixels from a set of coordinates.
zigzag_sample_image(h, w, scan_length, num_points, angle=np.pi / 20)
    Function for zigzag sampling an image.
zigzag_sample_surface(l, w, speed, sample_rate, time, angle=np.pi / 20)
    Function for zigzag sampling a surface.

Notes
-----
In principle, most of the scan pattern related parameters need only be
positive. However, it is assumed that the following requirements are fulfilled:

:Minimum length of scan area: 1 nm
:Minimum width of scan area: 1 nm
:Minimum scan speed: 1 nm/s
:Minimum sample_rate: 1 Hz
:Minimum scan time: 1 s
:Minimum scan length: 1 nm
:Minimum number of scan points: 1

Examples
--------
Sample a surface using a lissajous pattern:

>>> from magni.imaging.measurements import lissajous_sample_surface
>>> l = 13.0; w = 13.0; speed = 4.0; time = 27.0; sample_rate = 3.0;
>>> coords = lissajous_sample_surface(l, w, speed, sample_rate, time,
...                                   speed_mode=1)

Display the resulting pattern:

>>> from magni.imaging.measurements import plot_pattern
>>> plot_pattern(l, w, coords, 'surface')

Sample a 128x128 pixel image using a spiral pattern:

>>> from magni.imaging.measurements import spiral_sample_image
>>> h = 128; w = 128; scan_length = 1000.0; num_points = 200;
>>> coords = spiral_sample_image(h, w, scan_length, num_points)

Display the resulting pattern:

>>> plot_pattern(h, w, coords, 'image')

Find the corresponding unique pixels and plot the pixel mask:

>>> from magni.imaging.measurements import unique_pixels, plot_pixel_mask
>>> unique_pixels = unique_pixels(coords)
>>> plot_pixel_mask(h, w, unique_pixels)

"""

from magni.imaging.measurements._lissajous import *
from magni.imaging.measurements._matrices import *
from magni.imaging.measurements._random_line import *
from magni.imaging.measurements._spiral import *
from magni.imaging.measurements._square_spiral import *
from magni.imaging.measurements._uniform_line import *
from magni.imaging.measurements._uniform_rotated_line import *
from magni.imaging.measurements._util import *
from magni.imaging.measurements._visualisation import *
from magni.imaging.measurements._zigzag import *



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing unittests of the config modules.

Routine listings
----------------
UpdateTest(unittest.TestCase)
    Test if each config module can be updated with its defaults.

"""

from __future__ import division
import unittest

import magni


class UpdateTest(unittest.TestCase):
    """
    Test if each config module can be updated with its defaults.

    """

    def test_afm_config(self):
        config = magni.afm.config
        config.update(dict(config.items()))

    def test_cs_phase_transition_config(self):
        config = magni.cs.phase_transition.config
        config.update(dict(config.items()))

    def test_cs_reconstruction_amp_config(self):
        config = magni.cs.reconstruction.amp.config
        config.update(dict(config.items()))

    def test_cs_reconstruction_gamp_config(self):
        config = magni.cs.reconstruction.gamp.config
        config.update(dict(config.items()))

    def test_cs_reconstruction_iht_config(self):
        config = magni.cs.reconstruction.iht.config
        config.update(dict(config.items()))

    def test_cs_reconstruction_it_config(self):
        config = magni.cs.reconstruction.it.config
        config.update(dict(config.items()))

    def test_cs_reconstruction_sl0_config(self):
        config = magni.cs.reconstruction.sl0.config
        config.update(dict(config.items()))

    def test_utils_multiprocessing_config(self):
        config = magni.utils.multiprocessing.config
        config.update(dict(config.items()))



"""
..
    Copyright (c) 2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the core Approximate Message Passing (AMP) algorithm.

Routine listings
----------------
run(y, A)
    Run the AMP reconstruction algorithm.

See Also
--------
magni.cs.reconstruction.amp._config : Configuration options.
magni.cs.reconstruction.amp.threshold_operator : Threshold operators.
magni.cs.reconstruction.amp.stop_criterion : Stop criteria.
magni.cs.reconstruction.amp.util : Utilities.

Notes
-----
The default configuration of the AMP algorithm provides the Donoho, Maleki,
Montanari (DMM) AMP from [1]_ using the soft threshold with the "residual"
thresholding strategy from [2]_.

References
----------
.. [1] D. L. Donoho, A. Maleki, and A. Montanari, "Message-passing algorithms
   for compressed sensing", *Proceedings of the National Academy of Sciences of
   the United States of America*, vol. 106, no. 45, p. 18914-18919, Nov. 2009.
.. [2] A. Montanari, "Graphical models concepts in compressed sensing" *in
   Compressed Sensing: Theory and Applications*, Y. C. Eldar and G. Kutyniok
   (Ed.), Cambridge University Press, ch. 9, pp. 394-438, 2012.

"""

from __future__ import division
import copy

import numpy as np

from magni.cs.reconstruction.amp import config as _conf
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric


def run(y, A):
    """
    Run the AMP reconstruction algorithm.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray or magni.utils.matrices.{Matrix, MatrixCollection}
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.

    Returns
    -------
    alpha : ndarray
        The n x 1 reconstructed coefficient vector.
    history : dict, optional
        The dictionary of various measures tracked in the GAMP iterations.

    See Also
    --------
    magni.cs.reconstruction.amp._config : Configuration options.
    magni.cs.reconstruction.amp.threshold_operator : Threshold operators.
    magni.cs.reconstruction.amp.stop_criterion : Stop criteria.
    magni.cs.reconstruction.amp.util : Utilities.

    Notes
    -----
    Optionally, the algorithm may be configured to save and return the
    iteration history along with the reconstruction result. The returned
    history contains the following:

    * alpha_bar : Coefficient estimates (the reconstruction coefficients).
    * MSE : solution mean squared error (if the true solution is known).
    * threshold_parameters : The state of the threshold parameters.
    * stop_criterion : The currently used stop criterion.
    * stop_criterion_value : The value of the stop criterion.
    * stop_iteration : The iteration at which the algorithm stopped.
    * stop_reason : The reason for termination of the algorithm.

    Examples
    --------
    For example, recovering a vector from noiseless measurements using AMP
    with soft thresholding

    >>> import numpy as np
    >>> from magni.cs.reconstruction.amp import run, config
    >>> from magni.cs.reconstruction.amp.util import theta_mm
    >>> np.random.seed(seed=6028)
    >>> k, m, n = 10, 200, 400
    >>> A = 1 / np.sqrt(m) * np.random.randn(m, n)
    >>> alpha = np.zeros((n, 1))
    >>> alpha[:k] = np.random.normal(scale=2, size=(k, 1))
    >>> np_printoptions = np.get_printoptions()
    >>> np.set_printoptions(suppress=True, threshold=k+2)
    >>> alpha[:k + 2]
    array([[ 1.92709461],
           [ 0.74378508],
           [-3.2418159 ],
           [-1.32277347],
           [ 0.90118   ],
           [-0.19157262],
           [ 0.82855712],
           [ 0.24817994],
           [-1.43034777],
           [-0.21232344],
           [ 0.        ],
           [ 0.        ]])
    >>> y = A.dot(alpha)
    >>> threshold_params = {'threshold_level_update_method': 'residual',
    ... 'theta': theta_mm(float(m) / n), 'tau_hat_sq': 1.0}
    >>> config['threshold_parameters'] = threshold_params
    >>> alpha_hat = run(y, A)
    >>> alpha_hat[:k + 2]
    array([[ 1.92612707],
           [ 0.74185284],
           [-3.24179313],
           [-1.32200929],
           [ 0.90089208],
           [-0.19097308],
           [ 0.82658038],
           [ 0.24515825],
           [-1.42980997],
           [-0.2111469 ],
           [ 0.00002815],
           [ 0.00047293]])
    >>> np.sum(np.abs(alpha - alpha_hat) > 1e-2)
    0

    or recover the same vector returning a history comparing the pr. iteration
    solution to the true vector

    >>> config['report_history'] = True
    >>> config['true_solution'] = alpha
    >>> alpha_hat, history = run(y, A)
    >>> alpha_hat[:k + 2]
    array([[ 1.92612707],
           [ 0.74185284],
           [-3.24179313],
           [-1.32200929],
           [ 0.90089208],
           [-0.19097308],
           [ 0.82658038],
           [ 0.24515825],
           [-1.42980997],
           [-0.2111469 ],
           [ 0.00002815],
           [ 0.00047293]])
    >>> np.array(history['MSE']).reshape(-1, 1)[1:11]
    array([[ 0.01273323],
           [ 0.0053925 ],
           [ 0.00276334],
           [ 0.00093472],
           [ 0.0004473 ],
           [ 0.00021142],
           [ 0.00009618],
           [ 0.00004371],
           [ 0.00002275],
           [ 0.00001053]])

    or recover the same vector using the median based update method

    >>> config['report_history'] = False
    >>> threshold_params['threshold_level_update_method'] = 'median'
    >>> config['threshold_parameters'] = threshold_params
    >>> alpha_hat = run(y, A)
    >>> alpha_hat[:k + 2]
    array([[ 1.92577079],
           [ 0.7413161 ],
           [-3.24131125],
           [-1.32044146],
           [ 0.89998407],
           [-0.18981235],
           [ 0.82584016],
           [ 0.24445332],
           [-1.42934708],
           [-0.21043221],
           [ 0.        ],
           [ 0.        ]])
    >>> np.sum(np.abs(alpha - alpha_hat) > 1e-2)
    0

    >>> np.set_printoptions(**np_printoptions)

    """

    @_decorate_validation
    def validate_input():
        _numeric('y', ('integer', 'floating', 'complex'), shape=(-1, 1))
        _numeric('A', ('integer', 'floating', 'complex'), shape=(
            y.shape[0],
            _conf['true_solution'].shape[0]
            if _conf['true_solution'] is not None else -1))

    @_decorate_validation
    def validate_output():
        _numeric('alpha', ('integer', 'floating'), shape=(A.shape[1], 1))
        _generic('history', 'mapping',
                 keys_in=('alpha_bar', 'MSE', 'threshold_parameters',
                          'stop_criterion', 'stop_criterion_value',
                          'stop_iteration', 'stop_reason'))

    validate_input()

    # Initialisation
    init = _get_amp_initialisation(y, A)
    AH = init['AH']
    chi = init['chi']
    delta = init['delta']
    alpha_bar = init['alpha_bar']
    AH_dot_chi = init['AH_dot_chi']
    threshold = init['threshold']
    threshold_parameters = init['threshold_parameters']
    stop_criterion = init['stop_criterion']
    stop_criterion_name = init['stop_criterion_name']
    iterations = init['iterations']
    tolerance = init['tolerance']
    convert = init['convert']
    report_history = init['report_history']
    history = init['history']
    true_solution = init['true_solution']

    # AMP iterations
    for it in range(iterations):
        # Save previous state
        alpha_bar_prev = alpha_bar

        # AMP state updates
        # Basic AMP treshold operand: alpha_bar + AH_dot_chi
        alpha_bar = threshold.compute_threshold(locals())
        A_dot_alpha_bar = A.dot(alpha_bar)
        # Basic AMP deriv threshold operand: alpha_bar_prev + AH_dot_chi
        chi = y - A_dot_alpha_bar + 1 / delta * np.mean(
            threshold.compute_deriv_threshold(locals())) * chi
        AH_dot_chi = AH.dot(chi)

        # Threshold level update
        threshold.update_threshold_level(locals())

        # Stop criterion
        stop, stop_criterion_value = stop_criterion.compute(locals())

        # History reporting
        if report_history:
            history['alpha_bar'].append(alpha_bar)
            history['threshold_parameters'].append(
                copy.deepcopy(threshold.__dict__))
            history['stop_criterion_value'].append(stop_criterion_value)
            history['stop_iteration'] = it
            if true_solution is not None:
                history['MSE'].append(
                    1/A.shape[1] * np.linalg.norm(
                        true_solution - alpha_bar)**2)

        if stop:
            history['stop_reason'] = stop_criterion_name.upper()
            break

    alpha = alpha_bar

    validate_output()

    if report_history:
        return alpha, history
    else:
        return alpha


def _get_amp_initialisation(y, A):
    """
    Return an initialisation of the AMP algorithm.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray or magni.utils.matrices.{Matrix, MatrixCollection}
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.

    Returns
    -------
    init : dict
        The initialisation of the the AMP algorithm.

    """

    init = dict()
    param = dict(_conf.items())

    # Configured setup
    init['convert'] = param['precision_float']
    init['report_history'] = param['report_history']
    init['tolerance'] = param['tolerance']
    init['iterations'] = param['iterations']
    init['true_solution'] = param['true_solution']
    init['threshold_parameters'] = param['threshold_parameters']
    init['stop_criterion_name'] = param['stop_criterion'].__name__

    # Arguments based configuration
    init['AH'] = A.conj().T
    init['delta'] = A.shape[0] / A.shape[1]

    # threshold and stop criterion configuration
    threshold_init = copy.copy(init)
    threshold_init['A'] = A
    threshold_init['y'] = y
    init['threshold'] = param['threshold'](threshold_init)
    init['stop_criterion'] = param['stop_criterion'](threshold_init)

    # Configuration of remaining states
    if param['warm_start'] is not None:
        init['alpha_bar'] = init['convert'](param['warm_start'])
    else:
        init['alpha_bar'] = np.zeros((A.shape[1], 1), dtype=init['convert'])

    init['history'] = {'alpha_bar': [init['alpha_bar']],
                       'MSE': [np.nan],
                       'threshold_parameters': [
                           init['threshold_parameters']],
                       'stop_criterion': init['stop_criterion_name'].upper(),
                       'stop_criterion_value': [np.nan],
                       'stop_iteration': 0,
                       'stop_reason': 'MAX_ITERATIONS'}
    init['chi'] = np.zeros_like(y)  # by convention
    init['AH_dot_chi'] = init['AH'].dot(y - A.dot(init['alpha_bar']))

    return init



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing utility functions for initilisation of in- and output channels
in the Generalised Approximate Message Passing (GAMP) algorithm.

Routine listings
----------------
get_em_bg_amp_initialisation(problem_params, method='vila')
    Get initial parameters for EM Bernoulli-Guassian AMP.
rho_se(delta, zeta, resolution=1000)
    Return the theoretical noiseless LASSO phase transition.

"""

from __future__ import division

import numpy as np

from scipy.stats import norm as _gaus

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric
from magni.utils.matrices import norm as _norm


def get_em_bg_amp_initialisation(problem_params, method='vila'):
    """
    Get initial parameters for EM Bernoulli-Guassian AMP.

    If the initialisation `method` is `vila` then the scheme from [1]_ is used.
    If it is `krzakala` then the scheme from [2]_ is used.

    Parameters
    ----------
    problem_params : dict
        The problem parameters used to compute the initialisation.
    method : str
        The initialisation method to use.

    Returns
    -------
    tau : float
        The initial sparsity level.
    sigma_sq : float
        The initial AWGN noise level.
    theta_tilde : float
        The initial Gaussian prior variance.

    See Also
    --------
    magni.cs.reconstruction.gamp._output_channel.wrap_calculate_using_AWGN :\
        Related output channel.
    magni.cs.reconstruction.gamp._input_channels.wrap_calculate_using_iidsGB :\
        Related input channel.

    Notes
    -----
    Independently of the choice of `method`, the `problem_params` are:

    * y: the measurements
    * A: the system matrix

    If `method` is `vila`, one must also specify:

    * SNR: the signal-to-noise ratio

    References
    ----------
    .. [1] J. P. Vila and P. Schniter, "Expectation-Maximization
       Gaussian-Mixture Approximate Message Passing", *IEEE Transactions on
       Signal Processing*, 2013, vol. 61, no. 19, pp. 4658-4672, Oct. 2013.
    .. [2] F. Krzakala, M. Mezard, F. Sausset, Y. Sun, and L. Zdeborova,
       "Probabilistic reconstruction in compressed sensing: algorithms, phase
       diagrams, and threshold achieving matrices", *Journal of Statistical
       Mechanics: Theory and Experiment*, vol. P08009, pp. 1-57, Aug. 2012.

    Examples
    --------
    For example, get the "vila" initialisation for a SNR of 100

    >>> import numpy as np
    >>> from magni.cs.reconstruction.gamp.channel_initialisation import \
    ... get_em_bg_amp_initialisation as get_init
    >>> np.random.seed(6012)
    >>> A = np.random.randn(20, 40)
    >>> y = np.random.randn(20, 1)
    >>> problem_params = {'A': A, 'y': y, 'SNR': 100}
    >>> init_tup = get_init(problem_params, method='vila')
    >>> [round(float(elem), 3) for elem in init_tup]
    [0.193, 0.01, 0.123]

    or get the corresponding "krzakala" initialisation

    >>> del problem_params['SNR']
    >>> init_tup = get_init(problem_params, method='krzakala')
    >>> [round(float(elem), 3) for elem in init_tup]
    [0.05, 1.0, 0.479]

    """

    @_decorate_validation
    def validate_input():
        _generic('method', 'string', value_in=('vila', 'krzakala'))
        param_keys = {'y', 'A'}
        if method == 'vila':
            param_keys.add('SNR')
        _generic('problem_params', 'mapping', keys_in=tuple(param_keys))

    validate_input()

    y = problem_params['y']
    A = problem_params['A']
    m = A.shape[0]
    n = A.shape[1]
    delta = m / n

    norm_y_sq = np.linalg.norm(y)**2
    norm_A_sq = _norm(A, 'fro')**2

    if method == 'vila':
        SNR = problem_params['SNR']
        tau = delta * rho_se(delta, 2)
        sigma_sq = norm_y_sq / ((SNR + 1) * m)
        theta_tilde = (norm_y_sq - m * sigma_sq) / (norm_A_sq * tau)

    elif method == 'krzakala':
        tau = delta / 10
        sigma_sq = 1.0  # Our best guess. Does not seem to be documented in [2]
        theta_tilde = norm_y_sq / (norm_A_sq * tau)

    return tau, sigma_sq, theta_tilde


def rho_se(delta, zeta, resolution=1000):
    """
    Return the theoretical noiseless LASSO phase transition.

    Parameters
    ----------
    delta : float
        The under sampling ratio.
    zeta : {1, 2}
        The "problem" to get the phase transition for.
    resolution : int
        The resolution used in the brute force optimisation.

    Returns
    -------
    rho_se : float
        The phase transition value.

    Notes
    -----
    The theoretical noiseless LASSO phase transition is computed based on eq. 5
    in [3]_. A simple brute force optimisation with the specified `resolution`
    of that expression is used to find the phase transition. The "problems",
    for which the phase transition may be computed, are:

    1. Sparse nonnegative vectors
    2. Sparse signed vectors

    References
    ----------
    .. [3] D.L. Donoho, A. Maleki, and A. Montanari, "Message-passing
       algorithms for compressed sensing", *Proceedings of the National Academy
       of Sciences of the United States of America*, vol. 106, no. 45,
       pp. 18914-18919, Nov. 2009.

    Examples
    --------
    For example, find a phase transition value for a sparse signed vector:

    >>> from magni.cs.reconstruction.gamp.channel_initialisation import rho_se
    >>> round(float(rho_se(0.19, 2)), 3)
    0.238

    or find the corresponding value for a sparse nonnegative vector

    >>> round(float(rho_se(0.19, 1)), 3)
    0.318

    and a few more examples

    >>> round(float(rho_se(0.0, 1)), 3)
    0.0
    >>> round(float(rho_se(0.0, 2)), 3)
    0.0
    >>> round(float(rho_se(0.5, 1)), 3)
    0.558
    >>> round(float(rho_se(0.5, 2)), 3)
    0.386
    >>> round(float(rho_se(1.0, 1)), 3)
    0.95
    >>> round(float(rho_se(1.0, 2)), 3)
    0.95

    """

    @_decorate_validation
    def validate_input():
        _numeric('delta', ('integer', 'floating'), range_='[0;1]')
        _numeric('zeta', ('integer'), range_='[1;2]')
        _numeric('resolution', ('integer'), range_='[1;inf)')

    validate_input()

    gaus = _gaus()

    if delta < 1e-12:
        rho_se_z = 0.0
    elif delta > 0.99:
        rho_se_z = 0.95
    else:
        z = np.linspace(1e-12, 10, resolution)
        gaus_elem = (1 + z**2) * gaus.cdf(-z) - z * gaus.pdf(-z)
        rho_se_z = ((1 - zeta / delta * gaus_elem) /
                    (1 + z**2 - zeta * gaus_elem))

    return np.nanmax(rho_se_z)



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions for calculating the step-size (relaxation parameter)
used in the Iterative Thresholding algorithms.

Routine listings
----------------
calculate_using_adaptive(var)
    Calculate a step-size value in an 'adaptive' way.
calculate_using_fixed(var)
    Calculate the fixed step-size value.
get_function_handle(method)
    Return a function handle to a given calculation method.

"""

from __future__ import division

import numpy as np

from magni.utils.matrices import Matrix as _Matrix
from magni.utils.matrices import MatrixCollection as _MatrixC


def wrap_calculate_using_adaptive(var):
    """
    Arguments wrapper for culculate_using_adaptive.

    Calculate a step-size value in an 'adaptive' way.

    """

    convert = var['convert']
    A = var['A']

    def calculate_using_adaptive(var):
        """
        Calculate a step-size value in an 'adaptive' way.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in the calculation of the step-size.

        Returns
        -------
        kappa : float
            The step-size to be used in the IHT algorithm.

        Notes
        -----
        The 'adaptive' step-size selection from [1]_ is used.

        .. warning::

            This does not implement the Normalized IHT algorithm. It merely
            uses the adaptive choice of kappa for a 'correctly identified'
            support.

        References
        ----------
        .. [1] T. Blumensath and M.E. Davies, "Normalized Iterative Hard
           Thresholding: Guaranteed Stability and Performance", *IEEE Journal
           Selected Topics in Signal Processing*, vol. 4, no. 2, pp. 298-309,
           Apr. 2010.

        """

        support = var['alpha'] != 0

        if np.alltrue(~support):
            # For an empty support set, try with kappa=1
            g = np.ones(1)
            G = g

        elif isinstance(A, _Matrix) or isinstance(A, _MatrixC):
            g = var['c'].copy()
            g[~support] = 0
            G = A

        else:
            g = var['c'][support]
            G = A[:, support.ravel()]

        kappa = convert(np.linalg.norm(g)**2 / np.linalg.norm(G.dot(g))**2)

        return kappa

    return calculate_using_adaptive


def wrap_calculate_using_fixed(var):
    """
    Arguments wrapper for calculate_using_fixed.

    Calculate a fixed step-size value.

    """

    kappa = var['kappa']

    def calculate_using_fixed(var):
        """
        Calculate a fixed step-size value.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in the calculation of the step-size.

        Returns
        -------
        kappa : float
            The step-size to be used in the IHT algorithm.

        """

        return kappa

    return calculate_using_fixed


def get_function_handle(method, var):
    """
    Return a function handle to a given calculation method.

    Parameters
    ----------
    method : str
        Identifier of the calculation method to return a handle to.
    var : dict
        Local variables needed in the step-size method.

    Returns
    -------
    f_handle : function
        Handle to calculation `method` defined in this globals scope.

    """

    return globals()['wrap_calculate_using_' + method](var)



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions that return data about the system, magni, files, etc

Functions used to get various annotations and chases data are available from
this module.

See Also
--------
magni.reproducibility._annotation : Documentation of annotations
magni.reproducibility._chase : Documentation of chases

"""

from magni.reproducibility._annotation import *
from magni.reproducibility._chase import *



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions for evaluation of image reconstruction quality.

Routine listings
----------------
calculate_mse(x_org, x_recons)
    Function to calcualte Mean Squared Error (MSE).
calculate_psnr(x_org, x_recons, peak)
    Function to calculate Peak Signal to Noise Ratio (PSNR).
calculate_retained_energy(x_org, x_recons)
    Function to calculate the percentage of energy retained in reconstruction.

"""

from __future__ import division

import numpy as np

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


def calculate_mse(x_org, x_recons):
    r"""
    Calculate Mean Squared Error (MSE) between `x_recons` and `x_org`.

    Parameters
    ----------
    x_org : ndarray
        Array of original values.
    x_recons : ndarray
        Array of reconstruction values.

    Returns
    -------
    mse : float
        Mean Squared Error (MSE).

    Notes
    -----
    The Mean Squared Error (MSE) is calculated as:

    .. math::

         \frac{1}{N} \cdot \sum(x_{org} - x_{recons})^2

    where `N` is the number of entries in `x_org`.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.evaluation import calculate_mse
    >>> x_org = np.arange(4).reshape(2, 2)
    >>> x_recons = np.ones((2,2))
    >>> print('{:.2f}'.format(calculate_mse(x_org, x_recons)))
    1.50

    """

    @_decorate_validation
    def validate_input():
        _numeric('x_org', ('integer', 'floating'), shape=None)
        _numeric('x_recons', ('integer', 'floating'), shape=x_org.shape)

    validate_input()

    return ((x_org - x_recons) ** 2).mean()


def calculate_psnr(x_org, x_recons, peak):
    r"""
    Calculate Peak Signal to Noise Ratio (PSNR) between `x_recons` and `x_org`.

    Parameters
    ----------
    x_org : ndarray
        Array of original values.
    x_recons : ndarray
        Array of reconstruction values.
    peak : int or float
        Peak value.

    Returns
    -------
    psnr : float
        Peak Signal to Noise Ratio (PSNR) in dB.

    Notes
    -----
    The PSNR is as calculated as

    .. math::

         10 \cdot \log_{10}\left(\frac{peak^2}{ 1/N \cdot \sum(x_{org} -
         x_{recons})^2}\right)

    where `N` is the number of entries in `x_org`.

    If :math:`|x_{org} - x_{recons}| <= (10^{-8} + 1^{-5} * |x_{recons}|)`
    then `np.inf` is returned.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.evaluation import calculate_psnr
    >>> x_org = np.arange(4).reshape(2, 2)
    >>> x_recons = np.ones((2,2))
    >>> peak = 3
    >>> print('{:.2f}'.format(calculate_psnr(x_org, x_recons, peak)))
    7.78

    """

    @_decorate_validation
    def validate_input():
        _numeric('x_org', ('integer', 'floating'), shape=None)
        _numeric('x_recons', ('integer', 'floating'), shape=x_org.shape)
        _numeric('peak', ('integer', 'floating'), range_='(0;inf)')

    validate_input()

    if np.allclose(x_org, x_recons):
        psnr = np.inf
    else:
        psnr = 10 * np.log10(peak**2 / ((x_org - x_recons) ** 2).mean())

    return psnr


def calculate_retained_energy(x_org, x_recons):
    r"""
    Calculate percentage of energy retained in reconstruction.

    Parameters
    ----------
    x_org : ndarray
        Array of original values (must not be all zeros).
    x_recons : ndarray
        Array of reconstruction values.

    Returns
    -------
    energy : float
        Percentage of retained energy in reconstruction.

    Notes
    -----
    The retained energy is as calculated as

    .. math::
         \frac{\sum x_{recons}^2}{\sum x_{org}^2} \cdot 100\%

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.evaluation import calculate_retained_energy
    >>> x_org = np.arange(4).reshape(2, 2)
    >>> x_recons = np.ones((2,2))
    >>> print('{:.2f}'.format(calculate_retained_energy(x_org, x_recons)))
    28.57

    """

    @_decorate_validation
    def validate_input():
        _numeric('x_org', ('integer', 'floating'), shape=None)

        if np.count_nonzero(x_org) == 0:
            raise ValueError('x_org must not be all zeros')

        _numeric('x_recons', ('integer', 'floating'), shape=x_org.shape)

    validate_input()

    return (x_recons ** 2).sum() / (x_org ** 2).sum() * 100



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing input channel functions for the Generalised Approximate
Message Passing (GAMP) algorithm.

Routine listings
----------------
ValidatedMMSEInputChannel(magni.utils.validation.types.MMSEInputChannel)
    A base class for validated `magni.cs.reconstruction.gamp` input channels.
IIDBG(ValidatedMMSEInputChannel)
    An i.i.d. Bernoulli Gauss MMSE input channel.
IIDsGB(ValidatedMMSEInputChannel)
    An i.i.d. sparse Gauss Bernoulli MMSE input channel.

"""

from __future__ import division

import numpy as np

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric
from magni.utils.validation import validate_once as _validate_once
from magni.utils.validation.types import MMSEInputChannel as _MMSEInputChannel


class ValidatedMMSEInputChannel(_MMSEInputChannel):
    """
    A base class for validated `magni.cs.reconstruction.gamp` input channels.

    Parameters
    ----------
    var : dict
        The input channel state variables.

    """

    def __init__(self, var):
        super(ValidatedMMSEInputChannel, self).__init__(var)

        @_decorate_validation
        def validate_input():
            _generic('var', 'mapping')

        validate_input()

    @_validate_once
    def compute(self, var):
        """
        Compute the input channel value.

        Parameters
        ----------
        var : dict
            The variables used in computing of the input channel value.

        Returns
        -------
        mean : ndarray
            The computed input channel mean.
        variance : ndarray
            The computed input channel variance.

        Notes
        -----
        This method honors `magni.utils.validation.enable_allow_validate_once`.

        """

        @_decorate_validation
        def validate_input():
            _generic('var', 'mapping')

        validate_input()


class IIDBG(ValidatedMMSEInputChannel):
    """
    An i.i.d. Bernoulli Gauss MMSE input channel.

    This channel is based on equations (6), (7) in [3]_ and allows for using
    Expectation Maximization (EM) for learning the channel parameters as
    detailed in equations (19), (25), and (32) in [3]_ (see also [4]_ for an
    introduction to EM for GAMP).

    Parameters
    ----------
    tau : float or int
        The prior signal "density" (fraction of Gauss to Bernouilli).
    theta_bar : float or int
        The prior Gaussian mean.
    theta_tilde : float or int
        The prior Gaussian variance.
    use_em : bool
        The indicator of whether or not to use Expectation Maximazion (EM) to
        learn the prior parameters.

    Notes
    -----
    The above Parameters are the input channel parameters that must be passed
    in a `var` dict to the channel constructor.

    If `use_em` is True, the values given for `tau`, `theta_bar`, and
    `theta_tilde` are used for initialialisation.

    This channel is theoretically equivalent to the `IIDsGB` channel. However,
    due to numerical subtleties, it may give different results.

    References
    ----------
    .. [3] J. Vila, J. and P. Schniter, "Expectation-Maximization
       Bernoulli-Gaussian Approximate Message Passing", *in Forty Fifth
       Asilomar Conference on Signals, Systems and Computers (ASILOMAR)*,
       pp. 799-803, Pacific Grove, California, USA, Nov. 6-9, 2011
    .. [4] J. P. Vila and P. Schniter, "Expectation-Maximization
       Gaussian-Mixture Approximate Message Passing", *IEEE Transactions on
       Signal Processing*, 2013, vol. 61, no. 19, pp. 4658-4672, Oct. 2013.

    """

    def __init__(self, var):
        super(IIDBG, self).__init__(var)

        @_decorate_validation
        def validate_channel_parameters():
            _generic(('var', 'input_channel_parameters'), 'mapping')
            _numeric(('var', 'input_channel_parameters', 'tau'),
                     ('integer', 'floating'), range_='[0;1]')
            _numeric(('var', 'input_channel_parameters', 'theta_bar'),
                     ('integer', 'floating'))
            _numeric(('var', 'input_channel_parameters', 'theta_tilde'),
                     ('integer', 'floating'), range_='[0;inf)')
            _numeric(('var', 'input_channel_parameters', 'use_em'),
                     ('boolean'))

        validate_channel_parameters()

        c_params = var['input_channel_parameters']

        self.use_em = c_params['use_em']  # Whether or not to use EM learning
        self.tau = var['convert'](c_params['tau'])
        self.theta_bar = var['convert'](c_params['theta_bar'])
        self.theta_tilde = var['convert'](c_params['theta_tilde'])
        self.n = var['convert'](var['A'].shape[1])

    def compute(self, var):
        """
        Compute the IIDBG input channel value.

        Parameters
        ----------
        var : dict
            The variables used in computing of the input channel value.

        Returns
        -------
        mean : ndarray
            The computed input channel mean.
        variance : ndarray
            The computed input channel variance.

        """

        super(IIDBG, self).compute(var)

        s = var['s']
        r = var['r']
        tau = self.tau
        theta_bar = self.theta_bar
        theta_tilde = self.theta_tilde

        h_pzc = (theta_tilde * s) / (s + theta_tilde)
        g_pzc = (theta_bar / theta_tilde + r / s) * h_pzc
        k_pzc = 1 + (1 - tau) / tau * np.sqrt(theta_tilde / h_pzc) * np.exp(
            1/2 * ((r - theta_bar)**2 / (s + theta_tilde) - r**2 / s))
        l_pzc = 1.0 / k_pzc

        # New values of alpha_bar (mean) and alpha_tilde (variance)
        mean = l_pzc * g_pzc
        variance = l_pzc * (h_pzc + g_pzc**2) - l_pzc**2 * g_pzc**2

        if self.use_em:
            # EM update of tau
            # See Eq. (19) in [3]
            tau = 1.0 / self.n * np.sum(l_pzc)
            self.tau = var['convert'](tau)

            # EM update of theta_bar
            # See Eq. (25) in [3]
            theta_bar = 1.0 / (self.n * tau) * np.sum(mean)
            self.theta_bar = var['convert'](theta_bar)

            # EM update of theta_tilde
            # See Eq. (32) in [3]
            theta_tilde = 1.0 / (self.n * tau) * np.sum(
                l_pzc * ((theta_bar - g_pzc)**2 + h_pzc))
            self.theta_tilde = var['convert'](theta_tilde)

        return mean, variance


class IIDsGB(ValidatedMMSEInputChannel):
    """
    An i.i.d. sparse Gauss Bernoulli MMSE input channel.

    This channel is based on equations (68), (69) in [1]_ and allows for using
    Expectation Maximization (EM) for learning the channel parameters as
    detailed in equations (74), (78), and (79) in [1]_ (see also [2]_ for an
    introduction to EM for GAMP).

    Parameters
    ----------
    tau : float or int
        The prior signal "density" (fraction of Gauss to Bernouilli).
    theta_bar : float or int
        The prior Gaussian mean.
    theta_tilde : float or int
        The prior Gaussian variance.
    use_em : bool
        The indicator of whether or not to use Expectation Maximazion (EM) to
        learn the prior parameters.
    em_damping : float or int
        The damping of the EM updates (if using EM).

    Notes
    -----
    The above Parameters are the input channel parameters that must be passed
    in a `var` dict to the channel constructor.

    If `use_em` is True, the values given for `tau`, `theta_bar`, and
    `theta_tilde` are used for initialialisation. The `em_damping` must be in
    [0, 1) with 0 being no damping.

    This channel is theoretically equivalent to the `IIDBG` channel. However,
    due to numerical subtleties, it may give different results.

    References
    ----------
    .. [1] F. Krzakala, M. Mezard, F. Sausset, Y. Sun, and L. Zdeborova,
       "Probabilistic reconstruction in compressed sensing: algorithms, phase
       diagrams, and threshold achieving matrices", *Journal of Statistical
       Mechanics: Theory and Experiment*, vol. P08009, pp. 1-57, Aug. 2012.
    .. [2] J. P. Vila and P. Schniter, "Expectation-Maximization
       Gaussian-Mixture Approximate Message Passing", *IEEE Transactions on
       Signal Processing*, 2013, vol. 61, no. 19, pp. 4658-4672, Oct. 2013.

    """

    def __init__(self, var):
        super(IIDsGB, self).__init__(var)

        @_decorate_validation
        def validate_channel_parameters():
            _generic(('var', 'input_channel_parameters'), 'mapping')
            _numeric(('var', 'input_channel_parameters', 'tau'),
                     ('integer', 'floating'), range_='[0;1]')
            _numeric(('var', 'input_channel_parameters', 'theta_bar'),
                     ('integer', 'floating'))
            _numeric(('var', 'input_channel_parameters', 'theta_tilde'),
                     ('integer', 'floating'), range_='[0;inf)')
            _numeric(('var', 'input_channel_parameters', 'use_em'),
                     ('boolean'))

            if var['input_channel_parameters']['use_em']:
                _numeric(('var', 'input_channel_parameters', 'em_damping'),
                         ('integer', 'floating'), range_='[0;1)')

        validate_channel_parameters()

        c_params = var['input_channel_parameters']
        m = var['convert'](var['A'].shape[0])

        self.use_em = c_params['use_em']  # Whether or not to use EM learning
        if self.use_em:
            # EM damping level
            self.em_damp = var['convert'](c_params['em_damping'])
        self.tau = var['convert'](c_params['tau'])
        self.theta_bar = var['convert'](c_params['theta_bar'])
        self.theta_tilde = var['convert'](c_params['theta_tilde'])
        self.n = var['convert'](var['A'].shape[1])
        self.delta = var['convert'](m / self.n)

    def compute(self, var):
        """
        Compute the IIDsGB input channel value.

        Parameters
        ----------
        var : dict
            The variables used in computing of the input channel value.

        Returns
        -------
        mean : ndarray
            The computed input channel mean.
        variance : ndarray
            The computed input channel variance.

        """

        super(IIDsGB, self).compute(var)

        s = var['s']
        r = var['r']
        tau = self.tau
        theta_bar = self.theta_bar
        theta_tilde = self.theta_tilde

        q_denom = s + theta_tilde
        q_bar = (s * theta_bar + r * theta_tilde) / q_denom
        q_tilde = s * theta_tilde / q_denom

        fct1 = np.sqrt(1 + theta_tilde / s)
        arg1 = ((-r**2 * theta_tilde + s * theta_bar * (theta_bar - 2 * r)) /
                (2 * s * (s + theta_tilde)))
        fct2 = tau + (1 - tau) * fct1 * np.exp(arg1)

        # New values of alpha_bar (mean) and alpha_tilde (variance)
        mean = q_bar * tau / fct2
        variance = tau * (q_tilde + q_bar**2) / fct2 - mean**2

        if self.use_em:
            # EM update of tau
            # See Eq. (74) in [2]
            qem1 = 1/theta_tilde + 1/s
            qem2 = r / s + theta_bar / theta_tilde
            qem_exp = np.exp(
                1/2 * (qem2**2 / qem1 - theta_bar**2 / theta_tilde))
            qem_num = np.sum(qem1 / qem2 * mean)
            qem_denom = np.sum(
                1 / (1 - tau + tau * np.sqrt(qem1 / theta_tilde) * qem_exp))
            tau_f = qem_num / qem_denom
            if tau_f > self.delta:
                tau_f = self.delta
            tau = (1 - self.em_damp) * tau_f + self.em_damp * tau
            self.tau = var['convert'](tau)

            # EM update of theta_bar
            # See Eq. (78) in [2]
            theta_bar_f = np.sum(mean) / (tau * self.n)
            theta_bar = ((1 - self.em_damp) * theta_bar_f +
                         self.em_damp * theta_bar)
            self.theta_bar = var['convert'](theta_bar)

            # EM update of theta_tilde
            # See Eq. (79) in [2]
            theta_tilde_f = np.sum(
                variance + mean**2) / (tau * self.n) - theta_bar**2
            if theta_tilde_f < 0:
                theta_tilde_f = var['convert'](0)
            theta_tilde = ((1 - self.em_damp) * theta_tilde_f +
                           self.em_damp * theta_tilde)
            self.theta_tilde = var['convert'](theta_tilde)

        return mean, variance



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions for calculating a threshold (level) used in
Iterative Threshold algorithms.

Routine listings
----------------
calculate_far(delta)
    Calculate the optimal False Acceptance Rate for a given indeterminacy.
calculate_using_far(var)
    Calculate a threshold level using the FAR heuristic.
calculate_using_fixed(var)
    Calculate a threshold level using a given fixed support size.
get_function_handle(method)
    Return a function handle to a given calculation method.

"""

from __future__ import division

import numpy as np
import scipy.stats
try:
    import bottleneck as bn
    calculate_median = bn.median
    partsort = bn.partsort
except ImportError:
    calculate_median = np.median
    partsort = np.sort


def calculate_far(delta, it_algorithm):
    """
    Calculate the optimal False Acceptance Rate for a given indeterminacy.

    Parameters
    ----------
    delta : float
        The indeterminacy, m / n, of a system of equations of size m x n.
    it_algorithm : {IHT, ITS}
        The iterative thresholding algorithm to calculate the FAR for.

    Returns
    -------
    FAR : float
        The optimal False Acceptance Rate for the given indeterminacy.

    Notes
    -----
    The optimal False Acceptance Rate to be used in connection with the
    interference heuristic presented in the paper "Optimally Tuned Iterative
    Reconstruction Algorithms for Compressed Sensing" [1]_ is calculated from
    a set of optimal values presented in the same paper. The calculated value
    is found from a linear interpolation or extrapolation on the known set of
    optimal values.

    References
    ----------
    .. [1] A. Maleki and D.L. Donoho, "Optimally Tuned Iterative Reconstruction
       Algorithms for Compressed Sensing", *IEEE Journal Selected Topics in
       Signal Processing*, vol. 3, no. 2, pp. 330-341, Apr. 2010.

    """

    # Known optimal values (x - indeterminacy / y - FAR)
    x = [0.05, 0.11, 0.21, 0.41, 0.50, 0.60, 0.70, 0.80, 0.93]
    if it_algorithm == 'IHT':
        y = [0.0015, 0.002, 0.004, 0.011, 0.015, 0.02, 0.027, 0.035, 0.043]
    else:
        y = [0.02, 0.037, 0.07, 0.12, 0.16, 0.2, 0.25, 0.32, 0.37, 0.42]

    i = next((i for i in range(len(x) - 1) if delta <= x[i + 1]), len(x) - 2)

    FAR = y[i] + (delta - x[i]) * (y[i + 1] - y[i]) / (x[i + 1] - x[i])

    return FAR


def wrap_calculate_using_far(var):
    """
    Arguments wrapper for calculate_using_far.

    Calculate a threshold level using the FAR heuristic.

    """

    if 'hard' in var['threshold_alpha'].__name__:
        it_algorithm = 'IHT'
    else:
        it_algorithm = 'IST'

    far = calculate_far(var['A'].shape[0] / var['A'].shape[1],
                        it_algorithm)

    convert = var['convert']
    Lambda = convert(scipy.stats.norm.ppf(1 - far / 2))
    stdQ1 = convert(scipy.stats.norm.ppf(1 - 0.25))

    def calculate_using_far(var):
        """
        Calculate a threshold level using the FAR heuristic.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in the calculation of the threshold.

        Returns
        -------
        thres : float
            The threshold to be used in the Iterative Thresholding algorithm.

        Notes
        -----
        The threhold is calculated using a False Acceptance Ratio (FAR)
        heuristic as described in [1]_.

        References
        ----------
        .. [1] A. Maleki and D.L. Donoho, "Optimally Tuned Iterative
        Reconstruction Algorithms for Compressed Sensing", *IEEE Journal
        Selected Topics in Signal Processing*, vol. 3, no. 2, pp. 330-341,
        Apr. 2010.

        """

        c_median = calculate_median(np.abs(var['c'].ravel()))
        thres = var['kappa'] * Lambda * convert(c_median) / stdQ1

        return thres

    return calculate_using_far


def wrap_calculate_using_fixed(var):
    """
    Arguments wrapper for calculate_using_fixed.

    Calculate a threshold level using a given fixed support size.

    """

    k = var['param']['threshold_fixed']
    threshold_weights = var['threshold_weights']

    if partsort.__name__ == 'sort':
        # Fallback numpy sort
        def find_k_largest(coefs, k):
            return np.sort(coefs)[::-1][k]
    else:
        # Bottleneck partsort
        def find_k_largest(coefs, k):
            return partsort(coefs, len(coefs) - k)[len(coefs) - k - 1]

    def calculate_using_fixed(var):
        """
        Calculate a threshold level using a given fixed support size.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in the calculation of the threshold.

        Returns
        -------
        thres : float
            The threshold to be used in the Iterative Threshold algorithm.

        Notes
        -----
        The threshold is calculated using of a fixed support size i.e., by
        specifying the number of non-zero coefficients, k.

        """

        abs_coefficients = np.abs((var['alpha'] * threshold_weights).ravel())
        thres = find_k_largest(abs_coefficients, k)

        return thres

    return calculate_using_fixed


def get_function_handle(method, var):
    """
    Return a function handle to a given calculation method.

    Parameters
    ----------
    method : str
        Identifier of the calculation method to return a handle to.
    var : dict
        Local variables needed in the threshold method.

    Returns
    -------
    f_handle : function
        Handle to calculation `method` defined in this globals scope.

    """

    return globals()['wrap_calculate_using_' + method](var)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality for the analysis of dictionaries.

Routine listings
----------------
get_reconstructions(img, transform, fractions)
    Function to get a set of transform reconstructions.
show_coefficient_histogram(img, transforms, bins=None, range=None,
    output_path=None, fig_ext='pdf')
    Function to show a transform coefficient histogram.
show_psnr_energy_rolloff(img, reconstructions, fractions, return_vals=False,
    output_path=None, fig_ext='pdf')
    Function to show PSNR and retained energy rolloff of reconstructions.
show_reconstructions(coefficients, reconstructions, transform, fractions,
    output_path=None, fig_ext='pdf')
    Function to show transforms reconstructions and coefficients.
show_sorted_coefficients(img, transforms, output_path=None, fig_ext='pdf')
    Function to show a plot of transform coefficients in sorted order.
show_transform_coefficients(img, transforms, output_path=None, fig_ext='pdf')
    Function to show transform coefficients.
show_transform_quantiles(img, transform, fraction=1.0, area_mask=None)
    Function to show quantiles of transform coefficients.

"""

from __future__ import division
import os

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable as _make_axes_locatable

from magni.imaging import evaluation as _evaluation
from magni.imaging import visualisation as _visualisation
from magni.imaging import mat2vec as _mat2vec
from magni.imaging import vec2mat as _vec2mat
from magni.imaging.dictionaries import utils as _utils
from magni.imaging.visualisation import imshow as _imshow
from magni.imaging.visualisation import imsubplot as _imsubplot
from magni.reproducibility import io as _io
from magni.utils import split_path as _split_path
from magni.utils.multiprocessing import File as _File
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric
from magni.utils.validation import validate_levels as _levels


def get_reconstructions(img, transform, fractions):
    """
    Return transform reconstructions with different fractions of coefficents.

    The image `img` is transform coded using the specified `transform`.
    Reconstructions for the `fractions` of transform coefficients kept are
    returned along with the coefficients used in the reconstructions.

    Parameters
    ----------
    img : ndarray
        The image to get reconstructions of.
    transform : str
        The transform to use to obtain the reconstructions.
    fractions : list or tuple
        The fractions of coefficents to be used in the reconstructions.

    Returns
    -------
    coefficients : list
        The list of coefficients (each an ndarray) used in the reconstructions.
    reconstructions : list
        The list of reconstructions.

    Examples
    --------
    Get reconstructions from DCT based on 20% and 40% of the coefficients:

    >>> import numpy as np
    >>> from magni.imaging.dictionaries.analysis import get_reconstructions
    >>> img = np.arange(64).astype(np.float).reshape(8, 8)
    >>> transform = 'DCT'
    >>> fractions = (0.2, 0.4)
    >>> coefs, recons = get_reconstructions(img, transform, fractions)
    >>> len(recons)
    2
    >>> tuple(int(s) for s in coefs[0].shape)
    (8, 8)
    >>> tuple(int(s) for s in recons[0].shape)
    (8, 8)

    """

    @_decorate_validation
    def validate_input():
        _numeric('img', ('integer', 'floating', 'complex'), shape=(-1, -1))
        _generic('transform', 'string', value_in=_utils.get_transform_names())
        _levels('fractions', (_generic(None, 'explicit collection'),
                              _numeric(None, 'floating', range_='[0;1]')))

    @_decorate_validation
    def validate_output():
        _levels('coefficients',
                (_generic(None, 'explicit collection'),
                 _numeric(None, ('integer', 'floating', 'complex'),
                          shape=img.shape)))
        _levels('reconstructions',
                (_generic(None, 'explicit collection'),
                 _numeric(None, ('integer', 'floating', 'complex'),
                          shape=img.shape)))

    validate_input()

    coefficients = []
    reconstructions = []

    transform_matrix = _utils.get_function_handle(
        'matrix', transform)(img.shape)
    all_coefficients = _vec2mat(transform_matrix.conj().T.dot(_mat2vec(img)),
                                img.shape)
    sorted_coefficients = np.sort(np.abs(all_coefficients), axis=None)[::-1]

    for fraction in fractions:
        if int(fraction) == 1:
            used_coefficients = all_coefficients
        else:
            used_coefficients = np.zeros_like(all_coefficients)
            mask = np.abs(all_coefficients) > sorted_coefficients[
                int(np.round(fraction * all_coefficients.size)) - 1]
            used_coefficients[mask] = all_coefficients[mask]

        reconstruction = _vec2mat(
            transform_matrix.dot(_mat2vec(used_coefficients)).real, img.shape)

        coefficients.append(used_coefficients)
        reconstructions.append(reconstruction)

    validate_output()

    return coefficients, reconstructions


def show_coefficient_histogram(img, transforms, bins=None, range=None,
                               output_path=None, fig_ext='pdf'):
    """
    Show a histogram of coefficient values for different transforms.

    A histogram of the transform coefficient values for `img` using the
    different `transforms` are shown. If `output_path` is not None, the
    resulting figure and data used in the figure are saved.

    Parameters
    ----------
    img : ndarray
        The image to get transform coefficients for.
    transforms : list or tuple
        The names as strings of the transforms to use.
    bins : int
        The number of bins to use in the histogram (the default is None, which
        implies that the number of bins is determined based on the size of
        `img`).
    range : tuple
        The lower and upper range of the bins to use in the histogram (the
        default is None, which implies that the min and max values of `img`
        are used).
    output_path : str
        The output path (see notes below) to save the figure and data to (the
        default is None, which implies that the figure and data are not saved).
    fig_ext : str
        The figure extension determining the format of the saved figure (the
        default is 'pdf' which implies that the figure is saved as a PDF).

    See Also
    --------
    matplotlib.pyplot.hist : The underlying histogram plot function.

    Notes
    -----
    The `output_path` is specified as a path to a folder + an optional prefix
    to the file name. The remaining file name is fixed. If e.g, the fixed part
    of the file name was 'plot', then:

    * output_path = '/home/user/' would save the figure under /home/user/ with
      the name plot.pdf.
    * output_path = '/home/user/best' would save the figure under /home/user
      with the name best_plot.pdf.

    In addition to the saved figures, an annotated and chased HDF database with
    the data used to create the figures are also saved. The name of the HDF
    database is the same as for the figure with the exception that the file
    extension is '.hdf5'.

    Examples
    --------
    Save a coefficient histogram using the DCT and the DFT as transforms

    >>> import os, numpy as np
    >>> from magni.imaging.dictionaries import analysis as _a
    >>> img = np.arange(64).astype(np.float).reshape(8, 8)
    >>> transforms = ('DCT', 'DFT')
    >>> output_path = './histogram_test'
    >>> _a.show_coefficient_histogram(img, transforms, output_path=output_path)
    >>> current_dir = os.listdir('./')
    >>> for file in sorted(current_dir):
    ...     if 'histogram_test' in file:
    ...         print(file)
    histogram_test_coefficient_histogram.hdf5
    histogram_test_coefficient_histogram.pdf

    """

    @_decorate_validation
    def validate_input():
        _numeric('img', ('integer', 'floating', 'complex'), shape=(-1, -1))
        _levels('transforms',
                (_generic(None, 'explicit collection'),
                 _generic(None, 'string',
                          value_in=_utils.get_transform_names())))
        _numeric('bins', 'integer', range_='[1;inf)', ignore_none=True)
        _generic('output_path', 'string', ignore_none=True)
        _generic('fig_ext', 'string')
        # Range validated by matplotlib

    validate_input()

    fig, ax = plt.subplots(1, 1)

    datasets = dict()
    for transform in transforms:
        matrix_handle = _utils.get_function_handle('matrix', transform)
        coefficients = matrix_handle(img.shape).conj().T.dot(_mat2vec(img))

        if np.issubdtype(coefficients.dtype, np.complex):
            coefficients = np.abs(coefficients)

        label = '{} coefficients'.format(transform)
        alpha = 0.6
        edgecolor = 'none'

        if bins is not None:
            n, b, p = ax.hist(coefficients, bins=bins, range=range, log=True,
                              label=label, alpha=alpha, ec=edgecolor)
        else:
            n, b, p = ax.hist(coefficients, bins=10**int(np.log10(img.size)-1),
                              range=range, log=True, label=label, alpha=alpha,
                              ec=edgecolor)

        datasets[transform] = {'coefficients': coefficients, 'n': n, 'bins': b}

    leg = ax.legend()
    leg.get_frame().set_facecolor('1.0')
    ax.set_xlabel('Coefficient value (modulo for complex values)')
    ax.set_ylabel('Count')
    ax.set_title('Histogram of transform coefficients')
    for patch in leg.get_patches():
        patch.set_edgecolor(edgecolor)

    # Save figures and data
    if output_path is not None:
        _save_output(output_path, 'coefficient_histogram', fig, fig_ext,
                     datasets)


def show_psnr_energy_rolloff(img, reconstructions, fractions,
                             return_vals=False, output_path=None,
                             fig_ext='pdf'):
    """
    Show the PSNR and energy rolloff for the reconstructions.

    A plot of the Peak Signal to Noise Ratio (PSNR) and retained energy in the
    `recontructions` versus the `fractions` of coefficients used in the
    reconstructions is shown. If return_vals is True, the data used in the plot
    is returned. If `output_path` is not None, the resulting figure and data
    used in the figure are saved.

    Parameters
    ----------
    img : ndarray
        The image which the reconstructions are based on.
    reconstructions : list or tuple
        The reconstructions (each an ndarray) to show rolloff for.
    fractions : list or tuple
        The fractions of coefficents used in the reconstructions.
    return_vals : bool
        The flag indicating wheter or not to return the PSNR and energy values
        (the default is False, which indicate that the values are not
        returned).
    output_path : str
        The output path (see notes below) to save the figure and data to (the
        default is None, which implies that the figure and data are not saved).
    fig_ext : str
        The figure extension determining the format of the saved figure (the
        default is 'pdf' which implies that the figure is saved as a PDF).

    Returns
    -------
    psnrs : ndarray
       The PSNR values shown in the figure (only returned if return_vals=True).
    energy : ndarray
       The retained energy values shown in the figure (only returned if
       return_vals=True).

    Notes
    -----
    The `output_path` is specified as a path to a folder + an optional prefix
    to the file name. The remaining file name is fixed. If e.g, the fixed part
    of the file name was 'plot', then:

    * output_path = '/home/user/' would save the figure under /home/user/ with
      the name plot.pdf.
    * output_path = '/home/user/best' would save the figure under /home/user
      with the name best_plot.pdf.

    In addition to the saved figures, an annotated and chased HDF database with
    the data used to create the figures are also saved. The name of the HDF
    database is the same as for the figure with the exception that the file
    extension is '.hdf5'.

    Examples
    --------
    Save a PSNR and energy rolloff plot for reconstructions bases on the DCT:

    >>> import os, numpy as np
    >>> from magni.imaging.dictionaries import analysis as _a
    >>> img = np.arange(64).astype(np.float).reshape(8, 8)
    >>> transform = 'DCT'
    >>> fractions = (0.2, 0.4)
    >>> coefs, recons = _a.get_reconstructions(img, transform, fractions)
    >>> o_p = './rolloff_test'
    >>> _a.show_psnr_energy_rolloff(img, recons, fractions, output_path=o_p)
    >>> current_dir = os.listdir('./')
    >>> for file in sorted(current_dir):
    ...     if 'rolloff_test' in file:
    ...         print(file)
    rolloff_test_psnr_energy_rolloff.hdf5
    rolloff_test_psnr_energy_rolloff.pdf

    """

    @_decorate_validation
    def validate_input():
        _numeric('img', ('integer', 'floating', 'complex'), shape=(-1, -1))
        _levels('reconstructions',
                (_generic(None, 'explicit collection'),
                 _numeric(None, ('integer', 'floating', 'complex'),
                          shape=img.shape)))
        _levels('fractions', (_generic(None, 'explicit collection'),
                              _numeric(None, 'floating', range_='[0;1]')))
        _numeric('return_vals', 'boolean')
        _generic('output_path', 'string', ignore_none=True)
        _generic('fig_ext', 'string')

    validate_input()

    psnrs = np.zeros_like(fractions)
    energy = np.zeros_like(fractions)

    for k, reconstruction in enumerate(reconstructions):
        psnrs[k] = _evaluation.calculate_psnr(img, reconstruction,
                                              float(img.max()))
        energy[k] = _evaluation.calculate_retained_energy(img, reconstruction)

    fig, axes = plt.subplots(1, 2)
    axes[0].plot(fractions, psnrs)
    axes[0].set_xlabel('Fraction of coefficients')
    axes[0].set_ylabel('PSNR [dB]')
    axes[1].plot(fractions, energy)
    axes[1].set_xlabel('Fraction of coefficients')
    axes[1].set_ylabel('Retained energy [%]')
    fig.suptitle(('PSNR/Energy vs. fraction of coefficients used in ' +
                  'reconstruction'), fontsize=14)
    plt.tight_layout(rect=(0, 0, 1, 0.95))

    # Save figures and data
    if output_path is not None:
        datasets = {'psnrs': {'fractions': fractions, 'values': psnrs},
                    'energy': {'fractions': fractions, 'values': energy}}
        _save_output(output_path, 'psnr_energy_rolloff', fig, fig_ext,
                     datasets)

    # Return values
    if return_vals:
        return psnrs, energy


def show_reconstructions(coefficients, reconstructions, transform, fractions,
                         output_path=None, fig_ext='pdf'):
    """
    Show reconstructions and corresponding coefficients.

    Parameters
    ----------
    coefficients : list or tuple
        The coefficients (each an ndarray) used in the reconstructions.
    reconstructions : list or tuple
        The reconstructions (each an ndarray) to show.
    transform : str
        The transform used to obtain the reconstructions.
    fractions : list or tuple
        The fractions of coefficents used in the reconstructions.
    output_path : str
        The output path (see notes below) to save the figure and data to (the
        default is None, which implies that the figure and data are not saved).
    fig_ext : str
        The figure extension determining the format of the saved figure (the
        default is 'pdf' which implies that the figure is saved as a PDF).

    Notes
    -----
    The `output_path` is specified as a path to a folder + an optional prefix
    to the file name. The remaining file name is fixed. If e.g, the fixed part
    of the file name was 'plot', then:

    * output_path = '/home/user/' would save the figure under /home/user/ with
      the name plot.pdf.
    * output_path = '/home/user/best' would save the figure under /home/user
      with the name best_plot.pdf.

    In addition to the saved figures, an annotated and chased HDF database with
    the data used to create the figures are also saved. The name of the HDF
    database is the same as for the figure with the exception that the file
    extension is '.hdf5'.

    Examples
    --------
    Save images of coefficients and reconstructions based on the DCT:

    >>> import os, numpy as np
    >>> from magni.imaging.dictionaries import analysis as _a
    >>> img = np.arange(64).astype(np.float).reshape(8, 8)
    >>> trans = 'DCT'
    >>> fracs = (0.2, 0.4)
    >>> coefs, recons = _a.get_reconstructions(img, trans, fracs)
    >>> o_p = './reconstruction_test'
    >>> _a.show_reconstructions(coefs, recons, trans, fracs, output_path=o_p)
    >>> current_dir = os.listdir('./')
    >>> for file in sorted(current_dir):
    ...     if 'reconstruction_test' in file:
    ...         print(file)
    reconstruction_test_reconstruction_coefficients.hdf5
    reconstruction_test_reconstruction_coefficients.pdf
    reconstruction_test_reconstructions.hdf5
    reconstruction_test_reconstructions.pdf

    """

    @_decorate_validation
    def validate_input():
        _levels('coefficients',
                (_generic(None, 'explicit collection'),
                 _numeric(None, ('integer', 'floating', 'complex'),
                          shape=(-1, -1))))
        _levels('reconstructions',
                (_generic(None, 'explicit collection', len_=len(coefficients)),
                 _numeric(None, ('integer', 'floating', 'complex'),
                          shape=(-1, -1))))
        _generic('transform', 'string', value_in=_utils.get_transform_names())
        _levels('fractions', (_generic(None, 'explicit collection',
                                       len_=len(coefficients)),
                              _numeric(None, 'floating', range_='[0;1]')))
        _generic('output_path', 'string', ignore_none=True)
        _generic('fig_ext', 'string')

    validate_input()

    disp, axes_extent = _utils.get_function_handle(
        'visualisation', transform)(coefficients[0].shape)

    if len(coefficients) < 4:
        cols = 1
    else:
        cols = 4

    rows = int(np.ceil(len(coefficients) / cols))

    # Coefficients
    coef_imgs = [disp(_mat2vec(
        _visualisation.stretch_image(np.abs(coef), 1.0) + 1e-6))
        for coef in coefficients]
    labels = ['Frac: {:.3f}'.format(frac) for frac in fractions]
    fig_coef = _imsubplot(coef_imgs, rows, x_labels=labels, normalise=True)
    fig_coef.suptitle('Absolute value of transform coefficients used in ' +
                      'reconstructions (log-scale)', fontsize=14)

    # Reconstructions
    fig_recon = _imsubplot(reconstructions, rows, x_labels=labels,
                           normalise=True)
    fig_recon.suptitle('Reconstructions', fontsize=14)

    # Save figures
    if output_path is not None:
        fracs = ['frac' + str(fraction).replace('.', '')
                 for fraction in fractions]
        datasets_coef = {transform: {fraction: coefficients[k]
                                     for k, fraction in enumerate(fracs)}}
        datasets_recon = {transform: {fraction: reconstructions[k]
                                      for k, fraction in enumerate(fracs)}}
        _save_output(output_path, 'reconstruction_coefficients', fig_coef,
                     fig_ext, datasets_coef)
        _save_output(output_path, 'reconstructions', fig_recon, fig_ext,
                     datasets_recon)


def show_sorted_coefficients(img, transforms, output_path=None, fig_ext='pdf'):
    """
    Show the transform coefficient values in sorted order.

    A plot of the sorted coefficient values vs array index number is shown. If
    `output_path` is not None, the resulting figure and data used in the figure
    are saved.

    Parameters
    ----------
    img : ndarray
        The image to show the sorted transform coefficients values for.
    transforms : list or tuple
        The names as strings of the transforms to use.
    output_path : str
        The output path (see notes below) to save the figure and data to (the
        default is None, which implies that the figure and data are not saved).
    fig_ext : str
        The figure extension determining the format of the saved figure (the
        default is 'pdf' which implies that the figure is saved as a PDF).

    Notes
    -----
    The `output_path` is specified as a path to a folder + an optional prefix
    to the file name. The remaining file name is fixed. If e.g, the fixed part
    of the file name was 'plot', then:

    * output_path = '/home/user/' would save the figure under /home/user/ with
      the name plot.pdf.
    * output_path = '/home/user/best' would save the figure under /home/user
      with the name best_plot.pdf.

    In addition to the saved figures, an annotated and chased HDF database with
    the data used to create the figures are also saved. The name of the HDF
    database is the same as for the figure with the exception that the file
    extension is '.hdf5'.

    Examples
    --------
    Save a sorted transform coefficient plot for the DCT and DFT transforms:

    >>> import os, numpy as np
    >>> from magni.imaging.dictionaries import analysis as _a
    >>> img = np.arange(64).astype(np.float).reshape(8, 8)
    >>> transforms = ('DCT', 'DFT')
    >>> output_path = './sorted_test'
    >>> _a.show_sorted_coefficients(img, transforms, output_path=output_path)
    >>> current_dir = os.listdir('./')
    >>> for file in sorted(current_dir):
    ...     if 'sorted_test' in file:
    ...         print(file)
    sorted_test_sorted_coefficients.hdf5
    sorted_test_sorted_coefficients.pdf

    """

    @_decorate_validation
    def validate_input():
        _numeric('img', ('integer', 'floating', 'complex'), shape=(-1, -1))
        _levels('transforms',
                (_generic(None, 'explicit collection'),
                 _generic(None, 'string',
                          value_in=_utils.get_transform_names())))
        _generic('output_path', 'string', ignore_none=True)
        _generic('fig_ext', 'string')

    validate_input()

    fig, ax = plt.subplots(1, 1)

    datasets = dict()
    for transform in transforms:
        matrix_handle = _utils.get_function_handle('matrix', transform)
        coefficients = matrix_handle(img.shape).conj().T.dot(_mat2vec(img))
        sorted_coefficients = np.sort(np.abs(coefficients), axis=None)[::-1]

        ax.loglog(sorted_coefficients,
                  label='{} coefficients'.format(transform))

        leg = ax.legend()
        leg.get_frame().set_facecolor('1.0')
        ax.set_xlabel('Index value')
        ax.set_ylabel('Absolute coefficient value')
        ax.set_title('Absolute value of coefficients in sorted order')

        datasets[transform] = {'sorted_coefficients': sorted_coefficients}

    # Save figures and data
    if output_path is not None:
        _save_output(output_path, 'sorted_coefficients', fig, fig_ext,
                     datasets)


def show_transform_coefficients(img, transforms, output_path=None,
                                fig_ext='pdf'):
    """
    Show the transform coefficients.

    The transform coefficient of `img` are shown for the `transforms`. If
    `output_path` is not None, the resulting figure and data used in the figure
    are saved.

    Parameters
    ----------
    img : ndarray
        The image to show the transform coefficients for.
    transforms : list or tuple
        The names as strings of the transforms to use.
    output_path : str
        The output path (see notes below) to save the figure and data to (the
        default is None, which implies that the figure and data are not saved).
    fig_ext : str
        The figure extension determining the format of the saved figure (the
        default is 'pdf' which implies that the figure is saved as a PDF).

    Notes
    -----
    The `output_path` is specified as a path to a folder + an optional prefix
    to the file name. The remaining file name is fixed. If e.g, the fixed part
    of the file name was 'plot', then:

    * output_path = '/home/user/' would save the figure under /home/user/ with
      the name plot.pdf.
    * output_path = '/home/user/best' would save the figure under /home/user
      with the name best_plot.pdf.

    In addition to the saved figures, an annotated and chased HDF database with
    the data used to create the figures are also saved. The name of the HDF
    database is the same as for the figure with the exception that the file
    extension is '.hdf5'.

    Examples
    --------
    Save a figure showing the coefficients for the DCT and DFT transforms:

    >>> import os, numpy as np
    >>> from magni.imaging.dictionaries import analysis as _a
    >>> img = np.arange(64).astype(np.float).reshape(8, 8)
    >>> transforms = ('DCT', 'DFT')
    >>> o_p = './coefficient_test'
    >>> _a.show_transform_coefficients(img, transforms, output_path=o_p)
    >>> current_dir = os.listdir('./')
    >>> for file in sorted(current_dir):
    ...     if 'coefficient_test' in file:
    ...         print(file)
    coefficient_test_transform_coefficients.hdf5
    coefficient_test_transform_coefficients.pdf

    """

    @_decorate_validation
    def validate_input():
        _numeric('img', ('integer', 'floating', 'complex'), shape=(-1, -1))
        _levels('transforms',
                (_generic(None, 'explicit collection'),
                 _generic(None, 'string',
                          value_in=_utils.get_transform_names())))
        _generic('output_path', 'string', ignore_none=True)
        _generic('fig_ext', 'string')

    validate_input()

    if len(transforms) == 1:
        fig, axes = plt.subplots(1, 1, squeeze=False)
    else:
        rows = int(np.ceil(len(transforms) / 2))
        fig, axes = plt.subplots(rows, 2, squeeze=False)

    axes = axes.flatten()

    datasets = dict()
    for k, transform in enumerate(transforms):
        matrix_handle = _utils.get_function_handle('matrix', transform)
        visual_handle = _utils.get_function_handle('visualisation', transform)

        coefficients = matrix_handle(img.shape).conj().T.dot(_mat2vec(img))
        disp, axes_extent = visual_handle(img.shape)

        scaled_coefficients = _visualisation.stretch_image(
            np.abs(coefficients), 1.0) + 1e-6
        _imshow(disp(scaled_coefficients), ax=axes[k], show_axis='top',
                extent=axes_extent)
        axes[k].set_title(transform, y=1.05)

        datasets[transform] = {'coefficients': coefficients}

    # Save figures
    if output_path is not None:
        _save_output(output_path, 'transform_coefficients', fig, fig_ext,
                     datasets)


def show_transform_quantiles(img, transform, fraction=1.0, area_mask=None,
                             ax=None):
    """
    Show a plot of the quantiles of the transform coefficients.

    The `fraction` of `transform` coefficients holding the most energy for the
    image `img` is considered. The four quantiles within this fraction of
    coefficients are illustrated in the `transform` domain by showing
    coefficients between the different quantiles in different colours. If an
    `area_mask` is specified, only this area in the plot is higlighted whereas
    the rest is darkened.

    Parameters
    ----------
    img : ndarray
        The image to show the transform quantiles for.
    transform : str
        The transform to use.
    fraction : float
        The fraction of coefficients used in the quantiles calculation (the
        default value is 1.0, which implies that all coefficients are used).
    area_mask : ndarray
        Bool array of the same shape as `img` which indicates the area of the
        image to highlight (the default value is None, which implies that no
        particular part of the image is highlighted).
    ax : matplotlib.axes.Axes
        The axes on which the image is displayed (the default is None, which
        implies that a separate figure is created).

    Returns
    -------
    coef_count : dict
        Different counts of coeffcients within the `area_mask` (only returned
        if `area_mask` is not None).

    Notes
    -----
    The ticks on the colorbar shown below the figure are the percentiles of the
    entire set of coefficients corresponding to the quantiles with fraction of
    coefficients. For instance, if fraction is 0.10, then the percentiles are
    92.5, 95.0, 97.5, 100.0, corresponding to the four quantiles within the 10
    percent coefficients holding the most energy.

    The `coef_count` dictionary holds the following keys:

    * C_total : Total number of considered coefficients.
    * Q_potential : Number of potential coefficients within `mask_area`.
    * P_fraction : The fraction of Q_potential to the pixel count in `img`.
    * Q_total : Total number of (considered) coefficients within `mask_area`.
    * Q_fraction : The fraction of Q_total to Q_potential
    * QC_fraction : The fraction of Q_total to C_total.
    * Q0-Q1 : Number of coefficients smaller than the first quantile.
    * Q1-Q2 : Number of coefficients between the first and second quantile.
    * Q2-Q3 : Number of coefficients between the second and third quantile.
    * Q3-Q4 : Number of coefficients between the third and fourth quantile.

    Each of the QX-QY holds a tuple containing two values:

    1. The number of coefficients.
    2. The fraction of the number of coefficients to Q_total.

    Examples
    --------
    For example, show quantiles for a fraction of 0.2 of the DCT coefficients:

    >>> import numpy as np
    >>> from magni.imaging.dictionaries import analysis as _a
    >>> img = np.arange(64).astype(np.float).reshape(8, 8)
    >>> transforms = 'DCT'
    >>> fraction = 0.2
    >>> _a.show_transform_quantiles(img, transform, fraction=fraction)

    """

    @_decorate_validation
    def validate_input():
        _numeric('img', ('integer', 'floating', 'complex'), shape=(-1, -1))
        _generic('transform', 'string', value_in=_utils.get_transform_names())
        _numeric('fraction', 'floating', range_='[0;1]')
        _numeric('area_mask', 'boolean', shape=img.shape, ignore_none=True)
        _generic('ax', mpl.axes.Axes, ignore_none=True)

    @_decorate_validation
    def validate_output():
        _generic('coef_counts', 'mapping',
                 has_keys=('Q_total', 'Q_potential', 'Q0_Q1', 'Q_fraction',
                           'Q3_Q4', 'Q2_Q3', 'Q1_Q2', 'C_total',
                           'QC_fraction'))

    validate_input()

    # Colorbrewer qualitative 5-class Set 1 as colormap
    colours = [(228, 26, 28), (55, 126, 184), (77, 175, 74), (152, 78, 163),
               (255, 127, 0)]
    norm_colours = [tuple([round(val / 255, 4) for val in colour])
                    for colour in colours[::-1]]
    norm_colours = [norm_colours[0]] * 2 + norm_colours
    quantile_cmap = mpl.colors.ListedColormap(norm_colours)

    # Transform
    transform_matrix = _utils.get_function_handle(
        'matrix', transform)(img.shape)
    all_coefficients = _vec2mat(transform_matrix.conj().T.dot(_mat2vec(img)),
                                img.shape)
    # Force very low values to zero to avoid false visualisations
    all_coefficients[all_coefficients < np.finfo(np.float).eps * 10] = 0

    # Masked coefficients
    sorted_coefficients = np.sort(np.abs(all_coefficients), axis=None)[::-1]
    mask = np.abs(all_coefficients) > sorted_coefficients[
        int(np.round(fraction * all_coefficients.size)) - 1]

    used_coefficients = np.zeros_like(all_coefficients, dtype=np.float)
    used_coefficients[mask] = np.abs(all_coefficients[mask])

    # Quantiles
    q_linspace = np.linspace((1 - fraction) * 100, 100, 5)
    q_percentiles = tuple(q_linspace[1:4])
    quantiles = np.percentile(used_coefficients, q_percentiles)
    disp_coefficients = np.zeros_like(used_coefficients)
    disp_coefficients[(0 < used_coefficients) &
                      (used_coefficients <= quantiles[0])] = 1
    disp_coefficients[(quantiles[0] < used_coefficients) &
                      (used_coefficients <= quantiles[1])] = 2
    disp_coefficients[(quantiles[1] < used_coefficients) &
                      (used_coefficients <= quantiles[2])] = 3
    disp_coefficients[quantiles[2] < used_coefficients] = 4

    # Quantile figure
    disp, axes_extent = _utils.get_function_handle(
        'visualisation', transform)(img.shape)
    if ax is None:
        fig, axes = plt.subplots(1, 1)
    else:
        fig = ax.get_figure()
        axes = ax
    im = _imshow(disp(_mat2vec(10**disp_coefficients)), ax=axes,  # anti-log10
                 cmap=quantile_cmap, show_axis='top', interpolation='none',
                 extent=axes_extent)
    divider = _make_axes_locatable(axes)
    c_bar_ax = divider.append_axes('bottom', '5%', pad='3%')
    cbar = fig.colorbar(im, c_bar_ax, orientation='horizontal')
    cbar.solids.set_edgecolor("face")
    cbar.set_ticks([0.85, 1.705, 2.278, 2.85, 3.419, 4.0])
    cbar.set_ticklabels(['Excluded'] + [str(q) for q in q_linspace])

    plt.tight_layout(rect=(0, 0, 1, 0.95))

    # Area mask
    if area_mask is not None:
        _imshow(np.ma.array(np.ones_like(disp_coefficients), mask=area_mask),
                ax=axes, cmap='gray', show_axis='top', interpolation='none',
                extent=axes_extent, alpha=0.15)

        # Count of coefficients
        Q_total = np.sum(disp(_mat2vec(10**disp_coefficients))[area_mask] != 0)
        Qs = [np.sum(disp(_mat2vec(10**disp_coefficients))[area_mask] == k)
              for k in [1, 2, 3, 4]]
        coef_counts = {'Q' + str(k-1) + '_Q' + str(k):
                       (Qs[k-1], round(Qs[k-1] / Q_total, 2))
                       for k in [1, 2, 3, 4]}
        coef_counts['Q_total'] = Q_total
        coef_counts['Q_potential'] = np.sum(area_mask)
        coef_counts['Q_fraction'] = round(Q_total /
                                          coef_counts['Q_potential'], 2)
        coef_counts['C_total'] = np.sum(used_coefficients != 0)
        coef_counts['P_fraction'] = round(coef_counts['Q_potential'] /
                                          img.size, 2)
        coef_counts['QC_fraction'] = round(Q_total / coef_counts['C_total'], 2)

        validate_output()

        return coef_counts


def _save_output(output_path, name, fig, fig_ext, datasets):
    """
    Save figure and data output.

    Parameters
    ----------
    output_path : str
        The output_path to save to.
    name : str
        The 'fixed' part of the file name saved to.
    fig : matplotlib.figure.Figure
        The figure instance to save.
    fig_ext : str
        The file extension to use for the saved figure.
    datasets : dict
        The dict of dicts for datasets to save in a HDF database.

    """

    @_decorate_validation
    def validate_input():
        _generic('output_path', 'string')
        _generic('name', 'string')
        _generic('fig', mpl.figure.Figure)
        _generic('fig_ext', 'string')
        _levels('datasets', (_generic(None, 'mapping'),
                             _generic(None, 'mapping')))

    validate_input()

    if output_path[-1] == os.sep:
        path = output_path
        prefix = ''

    else:
        path, prefix, no_ext = _split_path(output_path)
        prefix = prefix + '_'

    fig.savefig(path + prefix + name + os.path.extsep + fig_ext)

    db_path = path + prefix + name + '.hdf5'
    _io.create_database(db_path)
    with _File(db_path, mode='a') as h5file:
        data_group = h5file.create_group('/', 'data', __name__ + ': ' + name)
        for dataset in datasets:
            set_group = h5file.create_group(data_group, dataset, dataset)
            for array in datasets[dataset]:
                h5file.create_array(set_group, array, datasets[dataset][array])
